import reveallib
import reveallib64
from utils import *
from multiprocessing.pool import Pool
import signal

def plotchains(ctg2mums,ctg2ref,contig2length,ref2length):

    for ctgname in ctg2ref:
        refs=set([refname for refname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci in ctg2ref[ctgname]])

        for refname in refs:
            # plt.ioff()
            plt.clf()
            plt.title(ctgname)
            plt.ylabel(ctgname)
            plt.xlabel(refname)

            #draw mums
            # mums=sorted(ctg2mums[ctgname][refname],key=lambda m: m[2],reverse=True)
            # if len(mums)>10000:
            #     logging.info("Too many mums, plot only the largest 10000")
            #     mums=mums[:10000]
            # for s1,s2,l,revcomp in mums:
            #     if revcomp==1:
            #         plt.plot([s1,s1+l],[s2+l,s2],'g-')
            #     else:
            #         plt.plot([s1,s1+l],[s2,s2+l],'r-')

            ax = plt.axes()
            # plt.xticks([], [])
            # plt.yticks([], [])
            last=0

            for ref,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci in sorted(ctg2ref[ctgname],key=lambda c: c[6] if c[1]==0 else c[5]):
                if refname != ref:
                    continue

                logging.info("Plot chain: %s"%str((ref,refname,revcomp,refbegin,refend,ctgbegin,ctgend,refend-refbegin)))

                if revcomp:
                    ctgbegin,ctgend=ctgend,ctgbegin

                # plt.axhline(y=ctgbegin,linewidth=.5,color='black',linestyle='solid')
                # plt.axhline(y=ctgend,linewidth=.5,color='black',linestyle='solid')
                # plt.axvline(x=refbegin,linewidth=.5,color='black',linestyle='solid')
                # plt.axvline(x=refend,linewidth=.5,color='black',linestyle='solid')

                # if last!=ctgbegin:
                #     ax.add_patch(
                #         patches.Rectangle(
                #             (0, last), #bottom left
                #             ref2length[refname], #width
                #             ctgbegin-last, #height
                #             alpha=.25,
                #             color="grey"
                #         )
                #     )

                ax.add_patch(
                    patches.Rectangle(
                        (refbegin, ctgbegin), #bottom left
                        refend-refbegin, #width
                        ctgend-ctgbegin, #height
                        alpha=.25,
                        color="green" if revcomp else "red"
                    )
                )

                if revcomp:
                    plt.plot([refbegin,refend],[ctgend,ctgbegin],'g--')
                else:
                    plt.plot([refbegin,refend],[ctgbegin,ctgend],'r--')

                last=ctgend

            plt.xlim(0,ref2length[refname])
            plt.ylim(0,contig2length[ctgname])
            # plt.savefig("chainlayout.svg")
            plt.show()
            # plt.close()

def plotclusters(ctg2mums,contig2length,ref2length):
    for ctg in ctg2mums:
        for ref in ctg2mums[ctg]:
            if len(ctg2mums[ctg][ref])>0:
                plt.clf()
                plt.title(ctg)
                plt.ylabel(ctg)
                plt.xlabel(ref)
                for refstart,ctgstart,cl,o in ctg2mums[ctg][ref]:
                    if o==1:
                        plt.plot([refstart,refstart+cl],[ctgstart+cl,ctgstart],'g-')
                    else:
                        plt.plot([refstart,refstart+cl],[ctgstart,ctgstart+cl],'r-')
                plt.xlim(0,ref2length[ref])
                plt.ylim(0,contig2length[ctg])
                plt.show()

def transform(args):

    try:
        from matplotlib import pyplot as plt
        from matplotlib import patches as patches
    except:
        pass

    logging.debug("Extracting mums.")

    if args.output==None:
        pref=[]
        for f in [os.path.basename(args.reference),os.path.basename(args.contigs)]:
            bn=os.path.basename(f)
            if '.' in bn:
                pref.append(bn[:bn.find('.')])
            else:
                pref.append(bn)
        args.output="_".join(pref)

    if args.nproc>1:
        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        pool = Pool(processes=2 if args.nproc>=2 else 1)
        signal.signal(signal.SIGINT, original_sigint_handler)
        try:
            async_result1 = pool.apply_async(getmums, (args.reference,args.contigs), {'sa64':args.sa64,'minlength':args.minlength,'cutN':args.cutn})
            async_result2 = pool.apply_async(getmums, (args.reference,args.contigs), {'revcomp':True,'sa64':args.sa64,'minlength':args.minlength,'cutN':args.cutn})
        except:
            pool.terminate()
            sys.exit(1)

        pool.close()
        pool.join()

        logging.debug("Retrieving results...")
        mums = async_result1.get()
        logging.debug("Done. MUMS in normal orientation: %d."%len(mums))

        logging.debug("Retrieving RC results...")
        rcmums = async_result2.get()
        logging.debug("Done. MUMS in reverse complemented orientation: %d."%len(rcmums))
    else:
        mums=getmums(args.reference,args.contigs,sa64=args.sa64,minlength=args.minlength,cutN=args.cutn)
        rcmums=getmums(args.reference,args.contigs,revcomp=True,sa64=args.sa64,minlength=args.minlength,cutN=args.cutn)

    reffile=os.path.basename(args.reference)
    ctgfile=os.path.basename(args.contigs)
    
    ref2length=dict()
    ref2seq=dict()
    for name,seq in fasta_reader(args.reference):
        ref2length[name]=len(seq)
        ref2seq[name]=seq
    
    contig2length=dict()
    contig2seq=dict()
    
    totl=0
    for name,seq in fasta_reader(args.contigs,cutN=args.cutn):
        l=len(seq)
        contig2length[name]=l
        totl+=l
        contig2seq[name]=seq

    #combine matches
    mums=mums+rcmums

    if len(mums)==0:
        logging.error("No mums! Exit")
        sys.exit()

    #if args.minlength==None:
    #    args.minlength=1
    
    if args.minlength==0: #auto determine minlength, prevent use of too many mums
        #sort by length
        logging.debug("Sorting %d MUMs by size..."%len(mums))
        mums=sorted(mums,key=lambda m: m[4],reverse=True)
        logging.debug("Done.")

        cov=0
        for i,mem in enumerate(mums):
            cov+=mem[4]
            if cov/float(totl)>1:
                break

        if i<len(mums)-1:
            mums=mums[:i+1]
            logging.info("Over representation of MUMs, auto determined min-mum-length to %d for cov. of %f"%(mums[-1][4],cov/float(totl)))

    ld=[mem[4] for mem in mums]
    bpcovered=sum(ld)

    bpncovered=totl-bpcovered
    if bpncovered<0:
        logging.info("Over representation of MUMs, probably better to use larger -m.")
        bpncovered=1
    
    avgcov=bpcovered/float(totl)

    if args.plot:
        from matplotlib import pyplot as plt
        from matplotlib import patches

    # if args.minchainsum==None: #auto set minchainsum with 0.5x of the genome wide coverage
        # args.minchainsum=int((.5*avgcov)*args.mineventsize)
        # logging.info("Auto determined minchainsum to %d"%args.minchainsum)

    logging.info("Assembly consists of %d contigs."%len(contig2seq))
    
    logging.debug("Associating mums to contigs.")
    #relate mums to contigs
    ctg2mums=mapmumstocontig(mums)

    logging.debug("Number of contigs that contain MUMs larger than %d: %d."%(args.minlength,len(ctg2mums)))
    
    logging.info("Cluster mums")
    ctg2mums=clustermumsbydiagonal(ctg2mums,maxdist=args.maxdist,minclustsize=args.mincluster)
    
    for i in range(args.extiter):
        logging.info("Extend cluster with local mums")
        ctg2mums=extend(ctg2mums,contig2seq,ref2seq,minlocallength=args.minlocallength,maxextend=args.maxextend)
        logging.info("Cluster again")
        ctg2mums=clustermumsbydiagonal(ctg2mums,maxdist=args.maxdist,minclustsize=args.mincluster)

    logging.info("Using %s to layout the assembly."%args.order)
    if args.order=='chains':
        ref2ctg,ctg2ref=chainstorefence(ctg2mums,contig2length,ref2length,maxmums=args.maxmums,mineventsize=args.mineventsize,minchainsum=args.minchainsum,nproc=args.nproc)
    else:
        ref2ctg,ctg2ref=contigstorefence(ctg2mums,contig2length,maxmums=args.maxmums,mineventsize=args.mineventsize,minchainsum=args.minchainsum,nproc=args.nproc)
    
    #write finished assembly based on contigs or chains that map on each reference chromosome
    if not args.split and args.outputtype=='fasta':
        finished=open(args.output+".fasta",'w')

    if not args.split and args.outputunmapped:
        unplaced=open(args.output+".unplaced.fasta",'w')
    
    totsequnplaced=0
    totseqplaced=0
    totseq=sum(contig2length.values())

    G=nx.MultiDiGraph()
    G.graph['paths']=[]
    G.graph['path2id']=dict()
    G.graph['id2path']=dict()
    G.graph['startnodes']=[]
    G.graph['endnodes']=[]
    
    gapi=0
    pathi=0
    
    defref2ctg=dict()
    unused=[]

    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    pool = Pool(processes=args.nproc)
    signal.signal(signal.SIGINT, original_sigint_handler)
    try:
        #multi-process contig-path computation
        for ref in ref2ctg:
            if ref=='unchained' or ref=='unplaced':
                defref2ctg[ref]=ref2ctg[ref]
                continue

            defref2ctg[ref]=pool.apply_async(bestctgpath, (ref2ctg[ref],))
    except KeyboardInterrupt:
        pool.terminate()
    else:
        pool.close()
    pool.join()

    #retrieve multi-process results
    for ref in ref2ctg:
        if ref=='unchained' or ref=='unplaced':
            continue
        
        b=set(ref2ctg[ref])
        defref2ctg[ref]=defref2ctg[ref].get()
        a=set(defref2ctg[ref])
        
        logging.debug("Selected %d out of %d %s to layout assembly with respect to %s."%(len(a),len(b),args.order,ref))
        
        if len(b)-len(a)>0:
            logging.debug("The following %d %s were placed on reference sequence %s but were not used in the layout:"%(len(b)-len(a),args.order,ref))
            if args.order=='contigs':
                for ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci in b - a:
                    logging.debug("Unused: %s (length=%d)"%(ctgname,contig2length[ctgname]))
                    ref2ctg['unplaced'].append(ctgname)
            else:
                for ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci in b - a:
                    if ctgbegin<ctgend:
                        logging.debug("Unused: (%s,%s,%s,%d) (length=%d)"%(ctgname,ctgbegin,ctgend,ci,ctgend-ctgbegin))
                        ref2ctg['unchained'][ctgname][ctgbegin:ctgend]=0
                    else:
                        logging.debug("Unused: (%s,%s,%s,%d) (length=%d)"%(ctgname,ctgbegin,ctgend,ci,ctgbegin-ctgend))
                        ref2ctg['unchained'][ctgname][ctgend:ctgbegin]=0
                    unused.append((ctgname,ci))

    # if args.plot:
    #     logging.debug("Plot chains before join.")
    #     plotchains(ctg2mums,ctg2ref,contig2length,ref2length)

    #remove unused chains from the ctg2ref mapping
    if args.order=="chains":
        defctg2ref=ctg2ref.copy()
        unused.sort(reverse=True)
        for name,i in unused:
            del defctg2ref[name][i]
            uchains=[]
            for chain in defctg2ref[name]:
                ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci=chain
                assert(ci!=i)
                if ci>i:
                    chain=ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci-1
                uchains.append(chain)
            defctg2ref[name]=uchains
        
        keys=sorted(defref2ctg)
        for ref in keys: #update the index of the chains, so that we can detect consecutive chains again
            if ref=='unchained' or ref=='unplaced':
            #     assert(False)
                continue
            for name,i in unused:
                ctgs=[]
                for ctg in defref2ctg[ref]:
                    ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci=ctg
                    assert(not(ctgname==name and ci==i))
                    if ctgname==name and ci>i:
                        ctg=ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci-1
                    ctgs.append(ctg)
                defref2ctg[ref]=ctgs

        logging.debug("Join consecutive chains")
        defref2ctg,defctg2ref=joinchains(defref2ctg,defctg2ref,ref2length,contig2length)

        #add parts of contigs that are not part of a chain
        logging.debug("Derive unchained sequence")
        addunchained(defref2ctg,defctg2ref,contig2length)

    else:
        defctg2ref=ctg2ref

    if args.plot:
        logging.debug("Plot chains after join.")
        plotchains(ctg2mums,defctg2ref,contig2length,ref2length)

    #build graph/fasta for the structural layout of the genome
    for ref in sorted(defref2ctg):
        
        pn=None

        if args.split and args.outputtype=='fasta':
            finished=open(args.output+"_"+ref.replace(" ","_").replace("|","").replace("/","").replace(";","").replace(":","")+".fasta",'w')
        
        if args.split and args.outputunmapped:
            unplaced=open(args.output+"_"+ref.replace(" ","_").replace("|","").replace("/","").replace(";","").replace(":","")+".unplaced.fasta",'w')
                
        if ref=='unchained' or ref=='unplaced':
            continue
        
        logging.info("Determining %s order for: %s"%(args.order,ref))
        
        ctgs=defref2ctg[ref]
        ctgs.sort(key=lambda c: c[3]) #sort by ref start position

        if args.plot:
            plt.clf()
            #plt.figure(0,figsize=(5,5))
            ax = plt.axes()
            plt.title(args.reference+" vs. "+args.contigs)
        
        coffset=0
        roffset=0
        
        yticks=[]
        yticklabels=[]
        base=os.path.splitext(os.path.basename(args.contigs))[0]
        ctgchromname=base+"_"+ref #name for the finished pseudomolecule
        ctgchromnameorg="*"+base+"_"+ref

        refid=len(G.graph['paths'])
        G.graph['path2id'][ctgchromname]=refid
        G.graph['id2path'][refid]=ctgchromname
        G.graph['paths'].append(ctgchromname)

        startnode=uuid.uuid4().hex
        G.add_node(startnode,offsets={refid:0},endpoint=True)
        G.graph['startnodes'].append(startnode)

        endnode=uuid.uuid4().hex
        G.add_node(endnode,offsets={refid:0},endpoint=True)
        G.graph['endnodes'].append(endnode)

        for ctg in ctgs:
            p="*"+base+"_"+ctg[0] #prefix with asterisk so they're recognisable
            if p not in G.graph['path2id']:
                G.graph['path2id'][p]=len(G.graph['paths'])
                G.graph['id2path'][len(G.graph['paths'])]=p
                G.graph['paths'].append(p)

        if args.outputtype=='fasta':
            finished.write(">%s (finished using %s)\n"%(ctgchromname,ref))
        
        i=0
        o=0

        refpath=[] #path that describes the 'transformed' genome
        orgpath=[] #path that describes the 'original' genome
        
        lastrefchain=False
        lastctgchain=False

        pctg=(None,ctgs[0][1],0,0,0,0,0,0,0)

        for ctg in ctgs:
            ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci=ctg

            if ctg==ctgs[-1]: #the last chain for this chromosome
                lastrefchain=True

            if (ci==len(defctg2ref[ctgname])-1 and not revcomp) or (ci==0 and revcomp): #the last chain for this contig 
                lastctgchain=True
            else:
                lastctgchain=False

            if revcomp:
                ctgbegin,ctgend=ctgend,ctgbegin
            
            pctgname,prevcomp,pscore,prefbegin,prefend,pctgbegin,pctgend,pctglength,pci=pctg
            
            if prevcomp:
                pctgbegin,pctgend=pctgend,pctgbegin
            
            if args.order=='chains':
                reloffset=ctgbegin
            else:
                reloffset=0
            
            if refend<=prefend:
                logging.error("Contained contig should not be in best contig path! %s with alignment length %d"%(ctgname,ctgend-ctgbegin))
                logging.error("pctg: %s"%str(pctg))
                logging.error("ctg: %s"%str(ctg))
                sys.exit(1)

            gapsize=refbegin-prefend
            
            if gapsize<0 or args.fixedsize:
                if gapsize<0:
                    logging.debug("Chains for contigs %s and %s overlap by %d bases."%(pctgname,ctgname,abs(gapsize)))
                gapsize=args.gapsize
            
            logging.debug("%s %d (index on ctg: %d->%d) - Order %s (revcomp=%d,prefstart=%d,prefend=%d,refstart=%d,refend=%d,ctgstart=%d,ctgend=%d,gapsize=%d)"%(ctgname,i,pci,ci,args.order,revcomp,prefbegin,prefend,refbegin,refend,ctgbegin,ctgend,gapsize))
            
            if args.order=='chains':
                event=None
                if ((ci==0 and revcomp==0) or (revcomp==1 and ci==len(defctg2ref[ctgname])-1)) and (pctgname==None or ((pci==len(ctg2ref[pctgname])-1 and prevcomp==0) or (prevcomp==1 and pci==0))): #consecutive contigs, no chains in between
                    event='contig break'
                else: #not first or last chain of contig, so has to be stuctural event
                    logging.debug("Non consecutive chains between %s [%d:%d:%d] and %s [%d:%d:%d]."%(pctgname,pctgbegin,pctgend,prevcomp,ctgname,ctgbegin,ctgend,revcomp))
                    if pctgname!=ctgname:
                        event="translocation between contigs" #between contig
                    else:
                        if revcomp!=prevcomp:
                            event="inversion"
                        else:
                            event="translocation within contig" #within contig
                    logging.info("Event of type: \'%s\' between %d and %d."%(event,prefend,refbegin))
                    logging.debug("Index within contig (%s, %d) layout: %d (of %d)"%(ctgname,revcomp,ci,len(defctg2ref[ctgname])))
                    if pctgname!=None:
                        logging.debug("Index within previous contig (%s, %d) layout: %d (of %d)"%(pctgname,prevcomp,pci,len(defctg2ref[pctgname])))
                
                logging.debug("Inserting gap of size: %d"%gapsize)
                
                gap=True
                
                if gapsize==0:
                    if args.outputtype=='fasta':
                        finished.write("N") #write at least one N so we can still distinguish events within fasta
                else:
                    if args.outputtype=='fasta':
                        finished.write("N"*gapsize)
                
                alength=ctgend-ctgbegin
                assert(alength>0)

                l=gapsize+alength
                
                if revcomp:
                    seq=rc(contig2seq[ctgname][ctgbegin:ctgend])
                else:
                    seq=contig2seq[ctgname][ctgbegin:ctgend]
                
                if args.outputtype=='fasta':
                    finished.write(seq)
                
                assert(ctgend-ctgbegin>=0)
                totseqplaced+=ctgend-ctgbegin
                assert(alength==ctgend-ctgbegin)
                
                if args.outputtype=='graph':
                    if event==None: #consecutive chains
                        G.node[pn]['seq']+=seq
                    else: #non-consecutive chains: different contig or structural variant
                        
                        if gapsize>0: #add a gap node
                            gapseq="N"*gapsize
                        else:
                            gapseq=""
                        
                        n=(ctgname,ctgbegin,ctgend,revcomp)
                        
                        G.add_node(n,seq=gapseq+seq,offsets={refid:o,G.graph['path2id']["*"+base+"_"+n[0]]:n[1]})

                        refpath.append(n)
                        
                        if pn!=None:
                            G.add_edge(pn,n,ofrom="+",oto="+",paths={refid})
                        else: #has to be first node for reference chrom
                            G.add_edge(startnode,n,ofrom="+",oto="+",paths={refid})

                        pn=n

                if args.plot:
                    
                    if gap:
                        ax.add_patch(
                            patches.Rectangle(
                                (0, o), #bottom left
                                ref2length[ref], #width
                                gapsize, #height
                                alpha=.25
                            )
                        )
                    
                    if revcomp:
                        ax.plot([refbegin,refend],[o+gapsize,o+alength+gapsize],'bx')
                        ax.plot([refbegin,refend],[o+gapsize,o+alength+gapsize],'g-')
                    else:
                        ax.plot([refbegin,refend],[o+gapsize,o+alength+gapsize],'bx')
                        ax.plot([refbegin,refend],[o+gapsize,o+alength+gapsize],'r-')
            
            else: # ordering contigs
                gap=True
                
                assert((pctglength-pctgend)>=0)

                alength=contig2length[ctgname]

                if prevcomp:
                    a_prefend=prefend+pctgbegin
                else:
                    a_prefend=prefend+(pctglength-pctgend)
                
                if revcomp:
                    a_refbegin=refbegin-(alength-ctgend)
                else:
                    a_refbegin=refbegin-ctgbegin

                gapsize=a_refbegin-a_prefend
                
                if gapsize==0: #perfect boundary, stil use one N to be able to distinguish the event
                    gapsize=1
                
                if gapsize<0 or args.fixedsize:
                    gapsize=args.gapsize
                
                if pctgname!=None:
                    logging.debug("\'%s\' follows \'%s\' inserting gap of size: %d"%(ctgname[:20],pctgname[:20],gapsize))
                    if args.outputtype=='fasta':
                        finished.write("N"*gapsize)
                
                assert(contig2length[ctgname]>=0)
                totseqplaced+=contig2length[ctgname]

                if revcomp:
                    seq=rc(contig2seq[ctgname])
                    if args.outputtype=='fasta':
                        finished.write(seq) #write the entire contig
                else:
                    seq=contig2seq[ctgname]
                    if args.outputtype=='fasta':
                        finished.write(seq)
                
                if args.outputtype=='graph':

                    gapi+=1
                    n=(gapi)
                    G.add_node(n,seq="N"*gapsize,offsets={refid:o})
                    if pn!=None:
                        G.add_edge(pn,n,ofrom="+",oto="+",paths={refid})
                    pn=n

                    n=(ctgname,0,contig2length[ctgname],revcomp)
                    G.add_node(n,seq=seq,offsets={refid:o+gapsize,G.graph['path2id']["*"+base+"_"+n[0]]:n[1]})
                    
                    if pn!=None:
                        G.add_edge(pn,n,ofrom="+",oto="+",paths={refid})
                    pn=n
                
                #l=gapsize+len(contig2seq[ctgname])
                assert(len(seq)==contig2length[ctgname])
                assert(gapsize>0)

                l=gapsize+contig2length[ctgname]
            
                if args.plot:
                    
                    if gap:
                        ax.add_patch(
                            patches.Rectangle(
                                (0, o), #bottom left
                                ref2length[ref], #width
                                gapsize, #height
                                alpha=.25
                            )
                        )
                    
                    if revcomp:
                        ax.plot([refbegin,refend],[o+gapsize+(alength-ctgend),o+(alength-ctgbegin)+gapsize],'bx')
                        ax.plot([refbegin,refend],[o+gapsize+(alength-ctgend),o+(alength-ctgbegin)+gapsize],'g-')
                    else:
                        ax.plot([refbegin,refend],[o+gapsize+ctgbegin,o+ctgend+gapsize],'bx')
                        ax.plot([refbegin,refend],[o+gapsize+ctgbegin,o+ctgend+gapsize],'r-')
            
            i+=1
            o=o+l
            yticks.append(o)
            yticklabels.append("%s:%d"%(ctgname[0:15],ctgend))
            # pctg=ctg

            if revcomp:
                pctg=ctgname,revcomp,score,refbegin,refend,ctgend,ctgbegin,ctglength,ci
            else:
                pctg=ctgname,revcomp,score,refbegin,refend,ctgbegin,ctgend,ctglength,ci

        if args.outputtype=='fasta':
            finished.write("\n")
        
        pathi+=2
        
        if args.split and args.outputtype=='fasta':
            finished.close()
        
        logging.debug("Done.")
        
        if args.plot:
            ax.set_yticks(yticks)
            ax.set_yticklabels(yticklabels)
            plt.xlim(0,ref2length[ref])
            plt.xlabel(ref)
            if args.interactive:
                plt.show()
            else:
                plt.savefig(args.output+"_"+ref.split()[0]+".png")

        G.add_edge(pn,endnode,ofrom="+",oto="+",paths={refid})

    if args.outputtype=='graph':
        ctgswithevents=set()

        if args.order=="chains":#reconnect the chains based on their layout in the draft assembly
            sortednodes=sorted([n for n in G.nodes() if type(n)!=str])

            if len(sortednodes)!=0:
                pn=sortednodes[0]

                startnode=uuid.uuid4().hex
                G.graph['startnodes'].append(startnode)
                G.add_node(startnode,offsets={G.graph['path2id']["*"+base+"_"+pn[0]]:0},endpoint=True)
                G.add_edge(startnode,pn,ofrom="+",oto="+" if pn[3]==0 else '-',paths={G.graph['path2id']["*"+base+"_"+pn[0]]})

                if len(sortednodes)>1:
                    for n in sortednodes[1:]:
                        if n[0]!=pn[0]:
                            startnode=uuid.uuid4().hex
                            G.graph['startnodes'].append(startnode)
                            G.add_node(startnode,offsets={G.graph['path2id']["*"+base+"_"+n[0]]:0},endpoint=True)
                            G.add_edge(startnode,n,ofrom="+",oto="+" if n[3]==0 else '-',paths={G.graph['path2id']["*"+base+"_"+n[0]]})

                            endnode=uuid.uuid4().hex
                            G.graph['endnodes'].append(endnode)
                            G.add_node(endnode,offsets={G.graph['path2id']["*"+base+"_"+pn[0]]:0},endpoint=True) #TODO: correct offset?
                            G.add_edge(pn,endnode,ofrom="+" if pn[3]==0 else '-',oto="+",paths={G.graph['path2id']["*"+base+"_"+pn[0]]})
                        else:
                            ctgswithevents.add("*"+base+"_"+pn[0])
                            G.add_edge(pn,n,ofrom="+" if pn[3]==0 else '-',oto="+" if n[3]==0 else '-',paths={G.graph['path2id']["*"+base+"_"+pn[0]]})
                        pn=n

                endnode=uuid.uuid4().hex
                G.graph['endnodes'].append(endnode)
                G.add_node(endnode,offsets={G.graph['path2id']["*"+base+"_"+pn[0]]:0},endpoint=True) #TODO: correct offset?
                G.add_edge(pn,endnode,ofrom="+" if pn[3]==0 else '-',oto="+",paths={G.graph['path2id']["*"+base+"_"+pn[0]]})

        if not args.allcontigs:
            G.graph['paths']=[sample for sample in G.graph['paths'] if sample in ctgswithevents or not sample.startswith("*")]

    if 'unplaced' in defref2ctg:
        if len(defref2ctg['unplaced'])>0:
            logging.info("The contigs could not be placed anywhere on the reference sequence.")
            for ctgname in defref2ctg['unplaced']:
                logging.info("%s length=%d"%(ctgname,contig2length[ctgname]))
                seq=contig2seq[ctgname]
                if args.outputunmapped:
                    unplaced.write(">%s\n"%(ctgname))
                    unplaced.write("%s\n"%seq)
                totsequnplaced+=len(seq)

    if 'unchained' in defref2ctg:
        if len(defref2ctg['unchained'])>0:
            logging.info("The following parts of contigs could not be placed anywhere on the reference sequence.")
            for name in defref2ctg['unchained']:
                # for start,end,i in defref2ctg['unchained'][name]:
                for start,end in defref2ctg['unchained'][name]:
                    logging.info("%s%s (start=%d,end=%d,length=%d,total-contig-length=%d)"%('*' if end-start!=contig2length[name] else '', name,start,end,end-start,contig2length[name]))
                    if args.outputunmapped:
                        unplaced.write(">%s[%d:%d]\n"%(name,start,end))
                        unplaced.write("%s\n"%contig2seq[name][start:end])
                    totsequnplaced+=end-start
    
    if not args.split and args.outputtype=='fasta':
        finished.close()
    
    if args.outputunmapped:
        unplaced.close()

    if args.outputtype=='graph':
        # write_gfa(G,None,outputfile=os.path.splitext(os.path.basename(args.contigs))[0],paths=True)
        write_gfa(G,None,outputfile=args.output,paths=True)
    
    if totseqplaced==0:
        logging.info("No sequence could be placed!")
    else:
        logging.info("%.2f%% (%d out of %d) of the assembly was placed with respect to the reference."% ( (totseqplaced/float(totseq))*100, totseqplaced, totseq ))

def addunchained(defref2ctg,defctg2ref,contig2length):
    #assign unchained parts
    defref2ctg['unchained']=dict()
    for ctg in contig2length: #assign (parts of) contigs that are not part of a chain
        # defref2ctg['unchained'][ctg]=IntervalTree()
        defref2ctg['unchained'][ctg]=[]
        offset=0
        if ctg in defctg2ref:
            defctg2ref[ctg].sort(key=lambda c: c[8])
            for ref,revcomp,score,refstart,refend,ctgstart,ctgend,l,ci in defctg2ref[ctg]:
                logging.debug("Checking domain %s:%d:%d %d."%(ctg,ctgstart,ctgend,ci))
                if revcomp:
                    ctgstart,ctgend=ctgend,ctgstart
                if ctgstart>offset:
                    logging.debug("Marking %s:%d:%d as unchained."%(ctg,offset,ctgstart))
                    # defref2ctg['unchained'][ctg][offset:ctgstart]=0
                    defref2ctg['unchained'][ctg].append((offset,ctgstart))
                offset=ctgend
        assert(offset<=contig2length[ctg])
        if offset<contig2length[ctg]:
            logging.debug("Marking %s:%d:%d as unchained."%(ctg,offset,contig2length[ctg]))
            defref2ctg['unchained'][ctg].append((offset,contig2length[ctg]))

def joinchains(ref2ctg,ctg2ref,ref2length,contig2length):
    # extref2ctg={'unchained':dict()}
    extref2ctg={}
    extctg2ref=dict()

    for ref in ref2ctg:
        if ref=="unchained":
            continue

        ref2ctg[ref]=sorted(ref2ctg[ref],key=lambda c: c[4])
        extref2ctg[ref]=[]

        pchain=None
        join=[]
        for ri,chain in enumerate(ref2ctg[ref]):
            # update=False
            ctgname,revcomp,score,refstart,refend,ctgstart,ctgend,l,ci=chain
            logging.debug("Evaluate chain: %s:%d:%d - %s:%d:%d %d %d"%(ctgname[:10],ctgstart,ctgend,ref[:10],refstart,refend,revcomp,ci))

            if len(extref2ctg[ref])>0:
                pctgname,prevcomp,pscore,prefstart,prefend,pctgstart,pctgend,pl,pci=extref2ctg[ref][-1]
                if pctgname==ctgname:
                    if revcomp==prevcomp:
                        if (not revcomp and ci==pci+1) or (revcomp and ci==pci-1): #consecutive chains, update boundaries
                            pctgname,prevcomp,pscore,prefstart,prefend,pctgstart,pctgend,pl,pci=extref2ctg[ref][-1]
                            logging.debug("Joining chains (%d): %d:%d - %d:%d --> %d:%d for contig: %s"%(revcomp,pctgstart,pctgend,ctgstart,ctgend,pctgstart,ctgend,ctgname))
                            prefend=refend
                            pctgend=ctgend
                            pscore+=score
                            extref2ctg[ref][-1]=(pctgname,prevcomp,pscore,prefstart,prefend,pctgstart,pctgend,pl,ci)
                            extctg2ref[ctgname][-1]=(ref,prevcomp,pscore,prefstart,prefend,pctgstart,pctgend,pl,ci)
                            continue
            
            extref2ctg[ref].append(chain)

            _,revcomp,score,refstart,refend,ctgstart,ctgend,l,ci=chain
            ctgchain=(ref,revcomp,score,refstart,refend,ctgstart,ctgend,l,ci)
            if ctgname not in extctg2ref:
                extctg2ref[ctgname]=[]
            extctg2ref[ctgname].append(ctgchain)

    return extref2ctg, extctg2ref

def decompose_contig(ctg,mums,contiglength,mineventsize=1500,minchainsum=1000,maxmums=15000):

    logging.debug("Determining best chain(s) for: %s"%ctg)
    paths=[]

    results=[]
    for ref in mums:
        rmums=mums[ref]

        candidatepaths=mempathsbothdirections(rmums,contiglength,n=maxmums,mineventsize=mineventsize,minchainsum=minchainsum)
        for path,score,rc,ctgstart,ctgend,refstart,refend in candidatepaths:
            if len(path)>0:
                paths.append((score,ctgstart,ctgend,refstart,refend,ref,rc,path))

    for r in results:
        candidatepaths=r.get()
        for path,score,rc,ctgstart,ctgend,refstart,refend in candidatepaths:
            if len(path)>0:
                paths.append((score,ctgstart,ctgend,refstart,refend,ref,rc,path))
    
    if len(paths)==0:
        return paths
    
    nrefchroms=len(set([p[5] for p in paths]))
    
    logging.debug("Found a total of %d chains for %s that map to %d different reference chromosomes."%(len(paths),ctg,nrefchroms))
    
    # return sorted(paths,key=lambda c: c[1] if c[6] else c[2])

    paths=sorted(paths,key=lambda c: c[0],reverse=True) #sort chains by alignment score
    
    selectedpaths=[]
    
    #take n-best paths that dont overlap on the query
    cit=IntervalTree()
    rit=IntervalTree()
    for path in paths:
        score,ctgstart,ctgend,refstart,refend,ref,revcomp,p=path
        
        logging.debug("Path before update mums: ctg:%d:%d - ref:%s:%d:%d (%d) with score %d"%(ctgstart,ctgend,ref,refstart,refend,revcomp,score))

        if revcomp:
            ctgend,ctgstart=ctgstart,ctgend

        np=[]
        for mum in p:
            for start,end,v in rit[mum[0]:mum[0]+mum[2]]:
                if start<=mum[0] and end>=mum[0]+mum[2]: #contained on ref domain
                    break
            else:
                for start,end,v in cit[mum[1]:mum[1]+mum[2]]:
                    if start<=mum[1] and end>=mum[1]+mum[2]: #contained on contig domain
                        break
                else:
                    np.append(mum)

        if len(np)==0:
            logging.debug("All mums are contained, skip")
            continue

        refstart=min([mum[0] for mum in np])
        refend=max([mum[0]+mum[2] for mum in np])
        ctgstart=min([mum[1] for mum in np])
        ctgend=max([mum[1]+mum[2] for mum in np])

        if revcomp:
            path=score,ctgend,ctgstart,refstart,refend,ref,revcomp,p
        else:
            path=score,ctgstart,ctgend,refstart,refend,ref,revcomp,p

        logging.debug("Path after update mums: ctg:%d:%d - ref:%s:%d:%d (%d) with score %d"%(ctgstart,ctgend,ref,refstart,refend,revcomp,score))
        assert(ctgstart<ctgend)

        s=cit[ctgstart:ctgend]
        sr=rit[refstart:refend]

        if s==set() and sr==set():
            cit[ctgstart:ctgend]=path
            rit[refstart:refend]=path
            selectedpaths.append(path)
        else:
            for start,end,v in s:
                if start<=ctgstart and end>=ctgend: #contained on contig domain
                    logging.debug("Chain: %s is contained on contig, skip it."%str(path))
                    break
            else:
                for start,end,v in sr:
                    if start<=refstart and end>=refend: #contained on reference domain
                        logging.debug("Chain: %s is contained on reference, skip it."%str(path))
                        break
                else:
                    if len(s)<=2 and len(sr)<=2:

                        for start,end,v in s:
                            if ctgstart<=start and ctgend>=end: #chain contains a smaller chain with better score, reduce to a point
                                logging.debug("Path %d-%d contains smaller, but better scoring chain %d-%d"%(ctgstart,ctgend,start,end))
                                ctgend=ctgstart
                                break
                            if ctgstart<=start: #left overlap, update ctgend
                                logging.debug("Update left overlap ctgend was %d, is %d"%(ctgend,start))
                                if revcomp:
                                    refstart+=ctgend-start
                                else:
                                    refend-=ctgend-start
                                ctgend=start
                            if ctgend>=end:
                                logging.debug("Update right overlap ctgstart was %d, is %d"%(ctgstart,end))
                                if revcomp:
                                    refend-=end-ctgstart
                                else:
                                    refstart+=end-ctgstart
                                ctgstart=end
                            
                            if ctgend-ctgstart<mineventsize:
                                break
                            if refend-refstart<mineventsize:
                                break
                        else:

                            logging.debug("Updated refstart=%d, refend=%d"%(refstart,refend))
                            assert(refend>=refstart)

                            sr=rit[refstart:refend]

                            for start,end,v in sr:
                                if refstart<=start and refend>=end: #chain contains a smaller chain with better score, reduce to a point
                                    logging.debug("Path %d-%d contains smaller, but better scoring chain %d-%d"%(refstart,refend,start,end))
                                    refend=refstart
                                    break
                                if refstart<=start: #left overlap, update ctgend
                                    logging.debug("Update left overlap refend was %d, is %d"%(refend,start))
                                    assert(refend-start>0)
                                    if revcomp:
                                        ctgstart+=refend-start
                                    else:
                                        ctgend-=refend-start
                                    refend=start
                                if refend>=end:
                                    logging.debug("Update right overlap refstart was %d, is %d"%(refstart,end))
                                    assert(end-refstart>0)
                                    if revcomp:
                                        ctgend-=end-refstart
                                    else:
                                        ctgstart+=end-refstart
                                    refstart=end
                                logging.debug("Updated ctgstart=%d, ctgend=%d."%(ctgstart,ctgend))

                                if ctgend-ctgstart<mineventsize:
                                    break
                                if refend-refstart<mineventsize:
                                    break
                            else:
                                
                                assert(ctgend>=ctgstart)

                                if ctgend>ctgstart and refend>refstart:
                                    if refend-refstart>mineventsize and ctgend-ctgstart>mineventsize:
                                        if revcomp:
                                            path=(score,ctgend,ctgstart,refstart,refend,ref,revcomp,p)
                                        else:
                                            path=(score,ctgstart,ctgend,refstart,refend,ref,revcomp,p)
                                        cit[ctgstart:ctgend]=path
                                        rit[refstart:refend]=path
                                        selectedpaths.append(path)
    
    for score,ctgstart,ctgend,refstart,refend,ref,revcomp,p in selectedpaths:
        logging.debug("Path after update: ctg:%d:%d - ref:%s:%d:%d (%d) with score %d"%(ctgstart,ctgend,ref,refstart,refend,revcomp,score))

    paths=sorted(selectedpaths,key=lambda c: c[1] if c[6] else c[2]) #sort by endposition on contig

    return paths

def chainstorefence(ctg2mums,contig2length,ref2length,mineventsize=1500,minchainsum=1000,maxmums=15000,nproc=1):    
    ref2ctg={'unchained':dict()}
    # ref2ctg={}
    ctg2ref=dict()
    results=dict()

    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    pool=Pool(processes=nproc)
    signal.signal(signal.SIGINT, original_sigint_handler)
    try:
        for ctg in ctg2mums:
            results[ctg]=pool.apply_async(decompose_contig,(ctg,ctg2mums[ctg],contig2length[ctg],),{'mineventsize':mineventsize,'minchainsum':minchainsum,'maxmums':maxmums})
    except KeyboardInterrupt:
        pool.terminate()
    else:
        pool.close()

    for ctg in ctg2mums:
        ref2ctg['unchained'][ctg]=IntervalTree()
        logging.debug("Determining best chains for: %s"%ctg)
        paths=results[ctg].get()
        
        if len(paths)==0:
            logging.info("No valid chains found for contig: %s"%ctg)
            ref2ctg['unchained'][ctg][0:contig2length[ctg]]=0
            continue
        
        logging.info("Found %d chains for contig: %s"%(len(paths),ctg))
        offset=0
        for i,path in enumerate(paths):
            logging.debug("Path %d: %s"%(i,str(path)))
            score,ctgstart,ctgend,refstart,refend,ref,revcomp,chain=path

            if revcomp:
                assert(ctgend<ctgstart)
            else:
                assert(ctgstart<ctgend)
            
            assert(offset<=ctgstart) #should not be any overlap on the contig anymore
            
            if ref in ref2ctg:
                ref2ctg[ref].append((ctg,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],i))
            else:
                ref2ctg[ref]=[(ctg,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],i)]
            
            if ctg in ctg2ref:
                ctg2ref[ctg].append((ref,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],i))
            else:
                ctg2ref[ctg]=[(ref,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],i)]
            
            if revcomp==1:
                ctgstart,ctgend=ctgend,ctgstart
            
            # if offset<ctgstart:
            if offset!=ctgstart:
                logging.debug("%d:%d:%d --> unchained"%(offset,ctgstart,revcomp))
                ref2ctg['unchained'][ctg][offset:ctgstart]=i
            logging.debug("%d:%d:%d --> %s:%d:%d"%(ctgstart,ctgend,revcomp,ref,refstart,refend))

            logging.debug("OFFSET:%d"%ctgend)

            offset=ctgend

        if offset!=contig2length[ctg]:
            assert(offset<contig2length[ctg])
            ref2ctg['unchained'][ctg][offset:contig2length[ctg]]=i

    return ref2ctg,ctg2ref

def map_contig(ctg,mums,contiglength,mineventsize=1500,minchainsum=1000,maxmums=15000):
    logging.debug("Determining best chain for: %s"%ctg)
    paths=[]
    for ref in mums:
        logging.debug("Checking %s"%ref)
        mpaths=mempathsbothdirections(mums[ref],contiglength,n=maxmums,all=False,mineventsize=mineventsize,minchainsum=minchainsum)
        if len(mpaths)>0:
            path,score,o,ctgstart,ctgend,refstart,refend=mpaths[0]
            paths.append((score,ctgstart,ctgend,refstart,refend,ref,o,path))
    return paths

def contigstorefence(ctg2mums,contig2length,mineventsize=1500,minchainsum=1000,maxmums=15000,nproc=1):
    
    ref2ctg={'unplaced':[]}
    ctg2ref=dict()
    results=dict()

    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    pool=Pool(processes=nproc)
    signal.signal(signal.SIGINT, original_sigint_handler)
    try:
        for ctg in ctg2mums:
            results[ctg]=pool.apply_async(map_contig,(ctg,ctg2mums[ctg],contig2length[ctg],),{'mineventsize':mineventsize,'minchainsum':minchainsum,'maxmums':maxmums})
    except KeyboardInterrupt:
        pool.terminate()
    else:
        pool.close()
    pool.join()

    for ctg in ctg2mums:
        paths=results[ctg].get()
        if len(paths)==0:
            ref2ctg['unplaced'].append(ctg)
            continue
        
        paths.sort(key=lambda p:p[0],reverse=True) #sort chains by score in descending order, best first
        
        score,ctgstart,ctgend,refstart,refend,ref,revcomp,chain=paths[0] #just take the best path
        
        if ref in ref2ctg:
            ref2ctg[ref].append((ctg,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],0))
        else:
            ref2ctg[ref]=[(ctg,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],0)]

        if ctg in ctg2ref:
            ctg2ref[ctg].append((ref,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],0))
        else:
            ctg2ref[ctg]=[(ref,revcomp,score,refstart,refend,ctgstart,ctgend,contig2length[ctg],0)]
    
    return ref2ctg,ctg2ref

def mapmumstocontig(mums):#,filtermums=False,mineventsize=1500,minchainsum=1000):
    ctg2mums=dict()
    logging.debug("Mapping %d mums to contigs."%len(mums))
    for mum in mums:
        refchrom, refstart, ctg, ctgstart, l, n, o = mum
        refstart=int(refstart)
        ctgstart=int(ctgstart)
        l=int(l)
        # n=int(n)
        o=int(o)
        if ctg in ctg2mums:
            if refchrom in ctg2mums[ctg]:
                ctg2mums[ctg][refchrom].append([refstart,ctgstart,l,o])
            else:
                ctg2mums[ctg][refchrom]=[[refstart,ctgstart,l,o]]

        else:
            ctg2mums[ctg]=dict({refchrom : [[refstart,ctgstart,l,o]]})
    
    return ctg2mums

def getmums(reference, query, revcomp=False, sa64=False, minlength=20, cutN=1000):
    if sa64:
        idx=reveallib64.index()
    else:
        idx=reveallib.index()
    
    t=IntervalTree()
    reffile=os.path.basename(reference)
    ctgfile=os.path.basename(query)

    idx.addsample(reffile)

    for name,seq in fasta_reader(reference):
        intv=idx.addsequence(seq)
        intv=Interval(intv[0],intv[1],name)
        t.add(intv)
    
    idx.addsample(ctgfile)

    for name,seq in fasta_reader(query,cutN=cutN):
        if revcomp:
            rcseq=rc(seq)
            intv=idx.addsequence(rcseq)
        else:
            intv=idx.addsequence(seq)
        intv=Interval(intv[0],intv[1],name)
        t.add(intv)
    
    idx.construct()
    
    mums=[]
    
    minlength=minlength if minlength!=0 else 1

    logging.debug("Extracting all MUMs of size %d or larger."%minlength)

    #TODO: remove intervaltree dependency here..

    for mum in idx.getmums(minlength):
        refstart=mum[1][0]#[1]
        ctgstart=mum[1][1]#[1]
        rnode=t[refstart].pop() #start position on match to node in graph
        cnode=t[ctgstart].pop()
        if revcomp:
            l=cnode[1]-cnode[0]
            mums.append((rnode[2], refstart-rnode[0], cnode[2], l-((ctgstart-cnode[0])+mum[0]), mum[0], mum[1], 1))
        else:
            mums.append((rnode[2], refstart-rnode[0], cnode[2], ctgstart-cnode[0], mum[0], mum[1], 0))

    if revcomp:
        logging.debug("Extracted %d, 3'-5' MUMs (size=%d bytes)."%(len(mums),sys.getsizeof(mums)))
    else:
        logging.debug("Extracted %d, 5'-3' MUMs (size=%d bytes)."%(len(mums),sys.getsizeof(mums)))
    
    return mums

def extend(ctg2mums,ctg2seq,ref2seq,maxextend=200,minlocallength=20):
    ctg2extmums=dict()
    for ctg in ctg2mums:
        ctg2extmums[ctg]=dict()
        for ref in ctg2mums[ctg]:
            ctg2extmums[ctg][ref]=list(ctg2mums[ctg][ref]) #copy

            for refstart,ctgstart,cl,o in ctg2mums[ctg][ref]:
                if o==0:
                    subref=ref2seq[ref][refstart-maxextend:refstart]
                    subctg=ctg2seq[ctg][ctgstart-maxextend:ctgstart]
                    if len(subctg)>0 and len(subref)>0:
                        idx=reveallib.index()
                        idx.addsample('ref')
                        idx.addsequence(subref)
                        idx.addsample('ctg')
                        idx.addsequence(subctg)
                        idx.construct()
                        for l,sps,r in idx.getmums(minlocallength):
                            relrefstart=sps[0]+refstart-len(subref)
                            relctgstart=sps[1]-(len(subref)+1)+ctgstart-len(subctg)
                            mum=(relrefstart, relctgstart, l, o)
                            ctg2extmums[ctg][ref].append(mum)

                    subref=ref2seq[ref][refstart+cl:refstart+cl+maxextend]
                    subctg=ctg2seq[ctg][ctgstart+cl:ctgstart+cl+maxextend]
                    if len(subctg)>0 and len(subref)>0:
                        idx=reveallib.index()
                        idx.addsample('ref')
                        idx.addsequence(subref)
                        idx.addsample('ctg')
                        idx.addsequence(subctg)
                        idx.construct()
                        for l,sps,r in idx.getmums(minlocallength):
                            relrefstart=sps[0]+refstart+cl
                            relctgstart=sps[1]+ctgstart+cl-(len(subref)+1)
                            mum=(relrefstart, relctgstart, l, o)
                            ctg2extmums[ctg][ref].append(mum)

                else: # o==1, reverse complement
                    subref=ref2seq[ref][refstart-maxextend:refstart] #extend pre
                    subctg=rc(ctg2seq[ctg][ctgstart+cl:ctgstart+cl+maxextend])
                    if len(subctg)>0 and len(subref)>0:
                        idx=reveallib.index()
                        idx.addsample('ref')
                        idx.addsequence(subref)
                        idx.addsample('ctg')
                        idx.addsequence(subctg)
                        idx.construct()
                        for l,sps,r in idx.getmums(minlocallength):
                            relrefstart=sps[0]+refstart-len(subref)
                            relctgstart=ctgstart+cl + (len(subctg)-(sps[1]-(len(subref)+1))-l)
                            mum=(relrefstart, relctgstart, l, o)
                            ctg2extmums[ctg][ref].append(mum)

                    subref=ref2seq[ref][refstart+cl:refstart+cl+maxextend] #extend suf
                    subctg=rc(ctg2seq[ctg][ctgstart-maxextend:ctgstart])
                    if len(subctg)>0 and len(subref)>0:
                        idx=reveallib.index()
                        idx.addsample('ref')
                        idx.addsequence(subref)
                        idx.addsample('ctg')
                        idx.addsequence(subctg)
                        idx.construct()
                        for l,sps,r in idx.getmums(minlocallength):
                            relrefstart=sps[0]+refstart+cl
                            relctgstart=(ctgstart-len(subctg)) + (len(subctg)-(sps[1]-(len(subref)+1))-l)
                            mum=(relrefstart, relctgstart, l, o)
                            ctg2extmums[ctg][ref].append(mum)

    return ctg2extmums

def clustermumsbydiagonal(ctg2mums,maxdist=90,minclustsize=65):
    before=0
    after=0

    ctg2clusters=dict()
    for ctg in ctg2mums:
        ctg2clusters[ctg]=dict()
        for ref in ctg2mums[ctg]:
            mums=ctg2mums[ctg][ref]
            before+=len(mums)
            
            rcmums=sorted([m for m in mums if m[3]==1],key=lambda m: (m[0]+(m[1]+m[2]), m[0]-(m[1]+m[2]))) #sort by anti-diagonal, then diagonal
            mums=sorted([m for m in mums if m[3]==0],key=lambda m: (  m[0]-m[1]     , m[0]+m[1])) #sort by diagonal, then anti-diagonal

            clusters=[]
            if len(mums)>0:
                pmum=mums[0]
                clusters=[pmum]
                for mum in mums[1:]: #cluster mums
                    if mum[0]-mum[1]==pmum[0]-pmum[1]: #same diagonal
                        
                        if mum[0]+mum[2]<pmum[0]+pmum[2]:
                            # print "mum",mum,"is contained in pmum",pmum,"on reference domain!"
                            continue

                        ddist=(mum[0]+mum[1])-(pmum[0]+pmum[2]+pmum[1]+pmum[2])
                        
                        if ddist < maxdist:
                            active=clusters[-1]
                            clusters[-1]=(active[0],active[1],(mum[0]+mum[2])-active[0],active[3])
                        else:
                            clusters.append(mum)
                    else:
                        clusters.append(mum)
                    pmum=mum

            rcclusters=[]
            if len(rcmums)>0:
                pmum=rcmums[0]
                rcclusters=[pmum]
                for mum in rcmums[1:]: #cluster mums

                    if (mum[0]+(mum[1]+mum[2]))==(pmum[0]+(pmum[1]+pmum[2])): #same anti-diagonal

                        if mum[0]+mum[2]<pmum[0]+pmum[2]:
                            # print "mum",mum,"is contained in pmum",pmum,"on reference domain!"
                            continue

                        ddist=(mum[0]-(mum[1]+mum[2])) - ((pmum[0]+pmum[2])-pmum[1])
                        
                        assert( (mum[0] - mum[1]+mum[2]) - (pmum[0]+pmum[2]-pmum[1]) >= 0)

                        if ddist < maxdist:
                            active=rcclusters[-1]
                            rcclusters[-1]=(active[0], mum[1], (mum[0]+mum[2])-active[0], active[3])
                        else:
                            rcclusters.append(mum)
                    else:
                        rcclusters.append(mum)

                    pmum=mum

            clusters=[c for c in clusters+rcclusters if c[2]>minclustsize]

            after+=len(clusters)

            ctg2clusters[ctg][ref]=clusters

    logging.info("Clustered %d mums into %d clusters."%(before,after))
    
    return ctg2clusters

def bestctgpath(chains):
    chains.sort(key=lambda c: (c[3],c[4])) #sort by reference 
    start=(0,0,0,0,0,0,0,0,0)
    
    link=dict()
    score=dict({start:0})
    
    processed=[]
    active=[start]
    maxscore=0
     
    for chain in chains:
        ctgname,revcomp,cscore,refbegin,refend,ctgbegin,ctgend,ctglength,ci=chain
        
        remove=[]
        for pctg in processed:
            pctgname,prevcomp,pscore,prefbegin,prefend,pctgbegin,pctgend,pctglength,pci=pctg

            if prefend<=refend: #may overlap, may not be contained
                active.append(pctg)
                remove.append(pctg)
        
        for r in remove:
            processed.remove(r)
        
        best=start
        w=0

        for actg in active:
            #calculate score of connecting to active point
            actgname,arevcomp,ascore,arefbegin,arefend,actgbegin,actgend,actglength,aci=actg
            
            if arefend>=refend:
                continue

            if arefend>refbegin:
                penalty=arefend-refbegin #penalize by the amount of overlap
            else:
                penalty=0
            
            tmpw=score[actg]+cscore-penalty
            if tmpw>w:
                w=tmpw
                best=actg
        
        assert(best!=None)
        
        link[chain]=best
        score[chain]=w
        
        if w>maxscore:
            maxscore=w
            end=chain
        
        processed.append(chain)
    
    #backtrack
    minscore=0
    path=[]
    while end[0]!=start[0]:
        path.append(end)
        end=link[end]

    return path[::-1]

def mempathsbothdirections(mums,ctglength,n=15000,mineventsize=1500,minchainsum=1000,wscore=1,wpen=1,all=True):
    nmums=len(mums)
    if nmums>n and n!=0: #take only n largest mums
        logging.info("Too many mums (%d), taking the %d largest."%(nmums,n))
        mums.sort(key=lambda mem: mem[2],reverse=True) #sort by size
        mums=mums[:n] #take the n largest
    
    if len(mums)==0:
        return []

    c=sum([m[2] for m in mums])
    logging.debug("Number of anchors: %d",len(mums))
    logging.debug("Sum of anchors: %d", c)
    logging.debug("Length of contig: %d", ctglength)
    logging.debug("Cov ratio: %s"% (c/float(ctglength)) )
    logging.debug("Min chain sum: %d"% minchainsum)
    logging.debug("Max gap size: %d"% mineventsize)
     
    paths=[]
    
    #extract the best path, remove mums that are part of or start/end within the range of the best path, until no more mums remain
    # while len(mums)>0:
        
    mums.sort(key=lambda mem: mem[0]+mem[2]) #sort by reference position
    
    init=(None, None, 0, 0, 0, 0)
    link=dict()
    score=dict({init:0})
    active=[]
    processed=[]
    start=init
    end=None
    
    endpoints=[]
    rcendpoints=[]
    pointtomem=dict()
    for mem in mums:
        # if mem[4]==0:
        if mem[3]==0:
            p=(mem[0]+mem[2],mem[1]+mem[2])
            endpoints.append(p)
            pointtomem[p]=mem
        else:
            p=(mem[0]+mem[2],mem[1])
            rcendpoints.append(p)
            pointtomem[p]=mem
    
    #build kdtrees
    memtree=kdtree(endpoints,2)
    rcmemtree=kdtree(rcendpoints,2)
    maxscore=0
    
    for mem in mums:
        best=init
        w=wscore*mem[2]
        
        # if mem[4]==1:
        if mem[3]==1:
            frompoint=(mem[0]-mineventsize, mem[1])
            topoint=(mem[0]+mem[2]-1, mem[1]+(mem[2]-1)+mineventsize)
            assert(frompoint[0]<topoint[0])
            assert(frompoint[1]<topoint[1])
            assert((topoint[1]-frompoint[1])==(topoint[0]-frompoint[0]))
            subactive=[pointtomem[p] for p in range_search(rcmemtree,frompoint,topoint)]
        else:
            frompoint=(mem[0]-mineventsize, mem[1]-mineventsize)
            topoint=(mem[0]+mem[2]-1, mem[1]+mem[2]-1)
            assert(frompoint[0]<topoint[0])
            assert(frompoint[1]<topoint[1])
            assert((topoint[1]-frompoint[1])==(topoint[0]-frompoint[0]))
            subactive=[pointtomem[p] for p in range_search(memtree,frompoint,topoint)]
        
        subactive.sort(key=lambda s: score[tuple(s)], reverse=True)

        for amem in subactive:
            if score[tuple(amem)]+(wscore*mem[2])<w:
                break

            #calculate score of connecting to active point
            if mem[3]==1:
                p1=(mem[0], mem[1]+mem[2])
                p2=(amem[0]+amem[2], amem[1])
                penalty=gapcost(p1,p2,lambda_=1,epsilon_=0,convex=True)
                assert(penalty>=0)
                tmpw=score[tuple(amem)]+(wscore*mem[2])-(wpen*penalty)
                if tmpw>w:
                    w=tmpw
                    best=amem
            else:
                p1=(amem[0]+amem[2], amem[1]+amem[2])
                p2=(mem[0], mem[1])
                penalty=gapcost(p1,p2,lambda_=1,epsilon_=0,convex=True)
                assert(penalty>=0)
                tmpw=score[tuple(amem)]+(wscore*mem[2])-(wpen*penalty)
                if tmpw>w:
                    w=tmpw
                    best=amem
        
        link[tuple(mem)]=tuple(best)
        score[tuple(mem)]=w
        
        if w>maxscore:
            maxscore=w
            end=tuple(mem)
    
    while len(link)!=0:
        path=[]
        o=end[3]

        while end!=start:
            tmp=tuple(end)
            assert(o==end[3])
            path.append(tmp)
            end=link[tmp]
            del link[tmp] #remove edge
            del score[tmp] #remove node

            if end not in link:
                break
        
        chainsum=sum([m[2] for m in path])
        
        if chainsum<minchainsum:
            break

        logging.info("Extracted path of length: %d (in mums) %d (sum of mums) %d (on ref in bp) with score: %s."%(len(path),chainsum,(path[0][0]+path[0][2])-path[-1][0], str(maxscore)) )
        
        # paths.append(path)
        refstart=path[-1][0]
        refend=path[0][0]+path[0][2]
        
        if o==1:
            ctgstart=path[-1][1]+path[-1][2]
            ctgend=path[0][1]
        else:
            ctgstart=path[-1][1]
            ctgend=path[0][1]+path[0][2]
        
        assert(maxscore<=chainsum)

        paths.append((path,maxscore,o,ctgstart,ctgend,refstart,refend))

        if not all: #just return first best path
            return paths

        mems=sorted([mem for mem in link],key=lambda m: m[0])
        
        maxscore=None
        score=dict()
        for mem in mems:
            if link[mem] not in score:
                score[mem]=mem[2]
                link[mem]=start
            else:
                score[mem]=mem[2]+score[link[mem]]

            if maxscore==None or score[mem]>maxscore:
                maxscore=score[mem]
                end=mem
    
    logging.info("Detected number of chains: %d."%len(paths))
    
    return paths



import networkx as nx
import logging
import gzip

#OVERRIDE intervaltree to bug in incorrect __eq__ function
import intervaltree
class IntervalPatched(intervaltree.Interval):    
    def __eq__(self,other):
        if type(self)==type(other)==Interval:
            return super(intervaltree.Interval,self).__eq__(other)
        else:
            return False
intervaltree.Interval=IntervalPatched
from intervaltree import Interval, IntervalTree


import sys
import os
from math import log
import uuid
import subprocess

def contract(G,topsort):
    newtopsort=[]
    stretches=[[]]
    pnode=topsort[0]
    newtopsort=[topsort[0]]
    for i,node in enumerate(topsort[1:]):
        
        assert(type(node)!=str)
        assert(type(pnode)!=str)

        pred=list(G.predecessors(node))
        suc=list(G.successors(pnode))

        if pred==[pnode] and suc==[node]:
            if len(stretches[-1])==0:
                stretches[-1].append(pnode)
            stretches[-1].append(node)
        else:
            if len(stretches[-1])!=0:
                stretches.append([])
            newtopsort.append(node)
        pnode=node

    for stretch in stretches:
        if len(stretch)>0:
            contract_nodes(G,stretch)

    assert(len(newtopsort)==len(set(newtopsort)))

    return newtopsort

def contract_nodes(G,nodes):
    logging.debug("Contract: Contracting nodes: %s"%nodes)
    G.node[nodes[0]]['seq']="".join([G.node[n]['seq'] for n in nodes])
    for n1,n2,data in G.out_edges(nodes[-1],data=True):
        G.add_edge(nodes[0],n2,**data)
    G.remove_nodes_from(nodes[1:])

def MultiGraphToDiGraph(G):
    structural_variants=[]
    toremove=[]
    logging.debug("Converting MultiDigraph to DiGraph, by removing structural variant edges.")
    orgpaths=set([G.graph['path2id'][p] for p in G.graph['paths'] if p.startswith('*')])
    refpaths=set([G.graph['path2id'][p] for p in G.graph['paths'] if not p.startswith('*')])
    refpathnames=[p for p in G.graph['paths'] if not p.startswith('*')]
    for e0,e1,k,d in G.edges(keys=True,data=True):
        if len(d['paths'] & refpaths)==0: #edge that exclusively represents a structural event 
            if type(e0)!=str and type(e1)!=str:
                if not G.has_edge(e1,e0): #the reverse edge does not exist
                    structural_variants.append((e0,e1,k,d))
            toremove.append((e0,e1))
    G.remove_edges_from(toremove)
    G.graph['paths']=refpathnames

    return structural_variants

def fasta_reader(fn,truncN=False,toupper=True,cutN=0,keepdash=False):
    seq=""
    gapseq=""
    sub=0

    if fn.endswith(".gz"):
        fopen=gzip.open
    else:
        fopen=open

    with fopen(fn,'r') as ff:
        for line in ff:
            line=line.rstrip()

            if line.startswith(">"):
                if seq!="":
                    if cutN>0:
                        yield name+"_"+str(sub),seq
                        sub=0
                    else:
                        yield name,seq
                name=line.replace(">","").replace("\t","")
                seq=""
            else:
                if toupper:
                    line=line.upper()
                    
                if not keepdash:
                    line=line.replace("-","")

                if truncN:
                    for base in line:
                        if base=='N':
                            if len(seq)==0:
                                continue
                            elif seq[-1]=='N':
                                continue
                            else:
                                seq+='N'
                        elif base=='-' and not keepdash:
                            pass
                        else:
                            seq+=base
                elif cutN>0:
                    for base in line:
                        if base=='N':
                            gapseq+='N'
                        else:
                            if len(gapseq)<cutN:
                                seq+=gapseq
                                gapseq=""
                            else:
                                if seq!="":
                                    yield name+"_"+str(sub),seq
                                    seq=""
                                    sub+=1
                                gapseq=""
                            seq+=base
                else:
                    seq+=line

        if seq!="":
            if cutN>0:
                yield name+"_"+str(sub),seq
            else:
                yield name,seq

def fasta_writer(fn,name_seq,lw=100):
    seq=""
    with open(fn,'w') as ff:
        for name,seq in name_seq:
            if not name.startswith(">"):
                name=">"+name+"\n"
            ff.write(name)
            for i in range( (len(seq)/lw)+(len(seq) % lw > 0)):
                ff.write(seq[i*lw:(i+1)*lw]+"\n")

def gapcost(pointa,pointb,model="sumofpairs",convex=False,lambda_=1,epsilon_=0): #epsilon is mismatch penalty weight, lambda is indel penalty weight
    assert(len(pointa)==len(pointb))
    
    if model=="star-avg":
        return abs(sum([pointa[i]-pointb[i] for i in range(len(pointa))]))/len(pointa)
    elif model=="star-med":
        return sorted([abs(pointa[i]-pointb[i]) for i in range(len(pointa))])[len(pointa)/2]
    elif model=="sumofpairs":
        p=0
        D=[abs(pointa[i]-pointb[i]) for i in range(len(pointa))]
        if epsilon_>0:
            p+=min(D)*epsilon_
        for i in range(len(D)): #all pairwise distances
            for j in range(i+1,len(D)):
                if convex:
                    p+=log(abs(D[i]-D[j])+1)*lambda_
                else:
                    p+=abs(D[i]-D[j])*lambda_
        return p
    else:
        logging.warn("Unknown penalty model: %s."%model)
        return 0

def rc(seq):
    d = {'A':'T','C':'G','G':'C','T':'A','N':'N','a':'t','c':'g',\
        'g':'c','t':'a','n':'n','Y':'R','R':'Y','K':'M','M':'K',\
        'S':'S','W':'W','B':'V','V':'B','D':'H','H':'D','N':'N',\
        'X':'X','-':'-'}
    return "".join([d[b] for b in reversed(seq)])

#extract all combinations for 'non-unique max exact matches' and return list of mems as if they were unique
def mem2mums(mem):
    import itertools
    l,n,spd=mem
    spd=sorted(spd, key=lambda m:m[1])
    pos=[[spd[0]]]
    for i in range(1,len(spd)):
        if spd[i-1][0]==spd[i][0]:
            pos[-1].append(spd[i])
        else:
            pos.append([spd[i]])
    mums=[]
    for t in itertools.product(*pos):
        yield (l,n,t)

def plotgraph(G, s1, s2, interactive=False, region=None, minlength=1):
    try:
        from matplotlib import pyplot as plt
        from matplotlib import patches as patches
    except:
        logging.error("Install matplotlib to generate mumplot.")
        return

    logging.debug("Generating plot for %s and %s."%(s1,s2))
    plt.xlabel(s1)
    plt.ylabel(s2)
    plt.title("REVEAL "+" ".join(sys.argv[1:]))
    maxx=0
    maxy=0
    
    minx=None
    miny=None

    #map names to ids
    s1=G.graph['path2id'][s1]
    s2=G.graph['path2id'][s2]

    logging.debug("Samples in graph: %s"%G.graph['path2id'])
    logging.debug("Generating plot for %s and %s, with minlength=%d."%(s1,s2,minlength))

    anchors=[]

    for node,data in G.nodes(data=True):
        if type(node)==str:
            continue

        if 'seq' in data:
            l=len(data['seq']) #either seq argument
        else:
            l=node.end-node.begin #or interval as node
        if l<minlength:
            continue
        
        s1t=False
        s2t=False

        if s1 in data['offsets']:
            s1t=True
            if minx==None:
                minx=data['offsets'][s1]
            if data['offsets'][s1]+l > maxx:
                maxx=data['offsets'][s1]+l
            if data['offsets'][s1] < minx:
                minx=data['offsets'][s1]
        else:
            continue

        if s2 in data['offsets']:
            s2t=True
            if miny==None:
                miny=data['offsets'][s2]
            if data['offsets'][s2]+l > maxy:
                maxy=data['offsets'][s2]+l
            if data['offsets'][s2] < miny:
                miny=data['offsets'][s2]
        else:
            continue

        if s1t and s2t:
            anchors.append((data['offsets'][s1],data['offsets'][s2],l))
    
    #plot all lines with one call to plot
    xlist,ylist= [],[]
    
    for s1,s2,l in anchors:
        xlist.append(s1)
        xlist.append(s1+l)
        xlist.append(None)
        ylist.append(s2)
        ylist.append(s2+l)
        ylist.append(None)

    plt.plot(xlist,ylist,'r-')

    if minx==None:
        minx=0
    if miny==None:
        miny=0
    
    plt.plot(minx,miny,'bx')
    plt.plot(maxx,maxy,'bx')
    
    if region!=None:
        rstart,rend=region.split(":")
        plt.axvline(x=int(rstart),linewidth=3,color='b',linestyle='dashed')
        plt.axvline(x=int(rend),linewidth=3,color='b',linestyle='dashed')
    
    if interactive:
        plt.show()
    else:
        plt.savefig("%s_%s.png"%(s1,s2))

def read_fasta(fasta, index, tree, graph, contigs=True, toupper=True):
    logging.info("Reading fasta: %s ..." % fasta)
    
    if 'paths' not in graph.graph:
        graph.graph['paths']=list()
    
    if 'id2path' not in graph.graph:
        graph.graph['id2path']=dict()
    
    if 'path2id' not in graph.graph:
        graph.graph['path2id']=dict()
    
    if 'id2end' not in graph.graph:
        graph.graph['id2end']=dict()

    if 'startnodes' not in graph.graph:
        graph.graph['startnodes']=[]

    if 'endnodes' not in graph.graph:
        graph.graph['endnodes']=[]

    if contigs:
        index.addsample(os.path.basename(fasta))
        for name,seq in fasta_reader(fasta,toupper=toupper):
            sid=len(graph.graph['paths'])
            name=name.replace(":","").replace(";","")
            if name in graph.graph['paths']:
                logging.fatal("Fasta with this name: \"%s\" is already contained in the graph."%name)
                sys.exit(1)
            graph.graph['paths'].append(name)
            graph.graph['path2id'][name]=sid
            graph.graph['id2path'][sid]=name
            graph.graph['id2end'][sid]=len(seq)

            intv=index.addsequence(seq)
            logging.debug("Adding interval: %s"%str(intv))
            Intv=Interval(intv[0],intv[1])
            tree.add(Intv)
            startnode=uuid.uuid4().hex
            endnode=uuid.uuid4().hex
            graph.add_node(startnode,offsets={sid:0},endpoint=True)
            graph.graph['startnodes'].append(startnode)
            graph.add_node(Intv,offsets={sid:0},aligned=0)
            graph.add_node(endnode,offsets={sid:len(seq)},endpoint=True)
            graph.graph['endnodes'].append(endnode)
            graph.add_edge(startnode,Intv,paths=set([sid]),ofrom='+',oto='+')
            graph.add_edge(Intv,endnode,paths=set([sid]),ofrom='+',oto='+')
    else: #treat every sequence in the multifasta as a target
        for name,seq in fasta_reader(fasta,toupper=toupper):
            index.addsample(name)
            sid=len(graph.graph['paths'])
            name=name.replace(":","").replace(";","")
            if name in graph.graph['paths']:
                logging.fatal("Fasta with this name: \"%s\" is already contained in the graph."%name)
                sys.exit(1)
            graph.graph['paths'].append(name)
            graph.graph['path2id'][name]=sid
            graph.graph['id2path'][sid]=name
            graph.graph['id2end'][sid]=len(seq)
            intv=index.addsequence(seq)
            logging.debug("Adding interval: %s"%str(intv))
            Intv=Interval(intv[0],intv[1])
            tree.add(Intv)
            startnode=uuid.uuid4().hex
            endnode=uuid.uuid4().hex
            graph.add_node(startnode,offsets={sid:0},endpoint=True)
            graph.graph['startnodes'].append(startnode)
            graph.add_node(Intv,offsets={sid:0},aligned=0)
            graph.add_node(endnode,offsets={sid:len(seq)},endpoint=True)
            graph.graph['endnodes'].append(endnode)
            graph.add_edge(startnode,Intv,paths=set([sid]),ofrom='+',oto='+')
            graph.add_edge(Intv,endnode,paths=set([sid]),ofrom='+',oto='+')

def read_gfa(gfafile, index, tree, graph, minsamples=1, maxsamples=None, targetsample=None, revcomp=False, remap=False):
    f=open(gfafile,'r')
    sep=";"
    nmapping={} #temp mapping object for nodeids in gfa file
    edges=[] #tmp list for edges
    paths=[]
    
    i=0

    if remap:
        gnodeid=graph.number_of_nodes()+1
    
    if 'paths' not in graph.graph:
        graph.graph['paths']=list()
    
    if 'id2path' not in graph.graph:
        graph.graph['id2path']=dict()
    
    if 'path2id' not in graph.graph:
        graph.graph['path2id']=dict()
    else:
        i=len(graph.graph['path2id'])
        assert(i not in graph.graph['id2path'])
    
    if 'id2end' not in graph.graph:
        graph.graph['id2end']=dict()

    if 'startnodes' not in graph.graph:
        graph.graph['startnodes']=[]

    if 'endnodes' not in graph.graph:
        graph.graph['endnodes']=[]

    for line in f:
        if line.startswith('H'):
            pass
        elif line.startswith('S'):
            s=line.strip().split('\t')
            nodeid=int(s[1])
            if len(s)==2: #node with empty sequence, allow for now
                s.append("")

            if remap:
                if graph.has_node(gnodeid):
                    logging.fatal("Id space for nodes is larger than total number of nodes in the graph.")
                    sys.exit(1)
            
            if index!=None:
                if revcomp:
                    intv=index.addsequence(rc(s[2]).upper())
                else:
                    intv=index.addsequence(s[2].upper())
                intv=Interval(intv[0],intv[1])
                tree.add(intv)
                graph.add_node(intv,aligned=0,offsets={})
                nmapping[nodeid]=intv
            else:
                if remap:
                    nmapping[nodeid]=gnodeid
                    gnodeid+=1
                else:
                    nmapping[nodeid]=nodeid

                if revcomp:
                    graph.add_node(nmapping[nodeid],seq=rc(s[2].upper()),aligned=0,offsets={})
                else:
                    graph.add_node(nmapping[nodeid],seq=s[2].upper(),aligned=0,offsets={})
        
        elif line.startswith('L'):
            edges.append(line)
        
        elif line.startswith('P'): #traverse paths to add offset values
            paths.append(line)

    for line in edges:
        e=line.strip().split("\t")

        if type(graph)==nx.DiGraph and (e[2]!='+' or e[4]!='+'):
            continue #skip these edges if we only want a directed acyclic graph

        #assert(not graph.has_edge(nmapping[int(e[1])],nmapping[int(e[3])]))
        #assert(not graph.has_edge(nmapping[int(e[3])],nmapping[int(e[1])]))
        tags=dict()
        tags['ofrom']=e[2]
        tags['oto']=e[4]

        if len(e)>5:
            tags['cigar']=e[5]
        
        if '*' in e: #there are additional tags parse them
            for tag in e[7:]:
                key,ttype,value=tag.split(':')
                tags[key.lower()]=value
        tags['paths']=set()

        graph.add_edge(nmapping[int(e[1])],nmapping[int(e[3])],**tags)

    if len(paths)==0:
        logging.fatal("No paths defined in GFA, exit.")
        sys.exit(1)

    if index==None:
        graph.graph['noffset']=max([v for v in nmapping.values() if type(v)==int])+1

    startnodes=set()
    endnodes=set()

    for line in paths:
        # logging.debug("Adding path...")
        cols=line.rstrip().split("\t")
        sample=cols[1]
        # logging.debug("Split done for %s."%sample)
        
        if type(graph)==nx.DiGraph:
            if sample.startswith("*"):
                logging.debug("DiGraph as input, so exclude path: %s"%sample)
                continue
        
        logging.debug("Adding path: %s"%sample)

        if sample in graph.graph['paths']:
            logging.fatal("ERROR: Graph already contains path for: %s"%sample)
            sys.exit(1)
        
        graph.graph['paths'].append(sample)
        
        if sample in graph.graph['path2id']:
            logging.fatal("ERROR: Graph already contains path for: %s"%sample)
            sys.exit(1)
        
        sid=len(graph.graph['path2id'])
        
        if sid in graph.graph['id2path']:
            logging.fatal("ERROR: Id %d already linked to a path in the graph."%sid)
            sys.exit(1)
        
        graph.graph['path2id'][sample]=sid
        graph.graph['id2path'][sid]=sample

        o=0
        
        # logging.debug("Split into tuple.")
        if len(cols)>=3:
            path=[(nid[:-1],nid[-1:]) for nid in cols[2].split(',')]
            # logging.debug("Done.")

            for pi,gfn in enumerate(path):
                nid,orientation=gfn
                node=nmapping[int(nid)]
                graph.node[node]['offsets'][sid]=o

                if 'seq' in graph.node[node]:
                    o+=len(graph.node[node]['seq'])
                elif isinstance(node,Interval):
                    o+=node[1]-node[0]
                else:
                    logging.warn("Node %s has unknown sequence content."%node)
                
                if pi==0:
                    pnode=node
                    pnid=nid
                    porientation=orientation
                    continue
                else:
                    if node not in graph[pnode]:
                        logging.fatal("Path %s has %s -> %s, but no edge between these nodes exists in the graph definition!"%(sample,pnid,nid))
                    assert(node in graph[pnode])
                    if type(graph)==nx.MultiDiGraph:
                        for i in graph[pnode][node]:
                            if graph[pnode][node][i]['oto']==orientation and graph[pnode][node][i]['ofrom']==porientation:
                                graph[pnode][node][i]['paths'].add(sid)
                                break
                        else:
                            logging.fatal("Edge missing for path %s between %s (%s) and %s (%s)"%(sample,pnode,porientation,node,orientation))
                            sys.exit(1)
                    else:
                        graph[pnode][node]['paths'].add(sid)

                pnode=node
                pnid=nid
                porientation=orientation
        
        # logging.debug("Loop over path done.")

        start=uuid.uuid4().hex
        graph.add_node(start,offsets={sid:0},endpoint=True)
        graph.add_edge(start,nmapping[int(path[0][0])],paths={sid},ofrom='+',oto=path[0][1])
        startnodes.add(start)

        end=uuid.uuid4().hex
        graph.add_node(end,offsets={sid:o},endpoint=True)
        graph.add_edge(nmapping[int(path[-1][0])],end,paths={sid},ofrom=path[-1][1],oto='+')
        endnodes.add(end)
        
        graph.graph['id2end'][sid]=o
    
    #remove nodes and edges that are not associated to any path
    remove=[]
    for n1,n2,d in graph.edges(data=True):
        if d['paths']==set(): #edge that is not traversed by any of the paths
            remove.append((n1,n2))
    if len(remove)>0:
        logging.info("Removing %d edges from the graph as they are not traversed..."%len(remove))
        for e in remove:
            logging.debug("%s"%str(e))
        graph.remove_edges_from(remove)
        logging.info("Done.")

    remove=[]
    for n,d in graph.nodes(data=True):
        if graph.node[n]['offsets']=={}: #node that is not traversed by any of the paths
            remove.append(n)
    if len(remove)>0:
        logging.info("Removing %d nodes from the graph as they are not traversed..."%len(remove))
        graph.remove_nodes_from(remove)
        logging.info("Done.")

    logging.debug("Extracting subgraphs...")
    conncomp=[comp for comp in nx.weakly_connected_components(graph)]
    logging.debug("Done.")

    #merge start/end nodes per connected component in the graph
    for i,comp in enumerate(conncomp):
        logging.debug("Inspecting connected component: %d (%d)"%(i,len(comp)))

        startmerge=set()
        endmerge=set()
        for node in comp:
            if node in startnodes:
                startmerge.add(node)
            if node in endnodes:
                endmerge.add(node)

        if len(endmerge)>0:
            endnode=uuid.uuid4().hex            
            graph.add_node(endnode,offsets={},seq="",endpoint=True) #add dummy node for end of each sequence in the subgraph
            graph.graph['endnodes'].append(endnode)

            for node in endmerge: #copy offset values
                for k in graph.node[node]['offsets']:
                    graph.node[endnode]['offsets'][k]=graph.node[node]['offsets'][k]
                #reconnect
                pred=set()
                predids=set()
                for pnode in graph.predecessors(node):
                    if type(graph)==nx.MultiDiGraph:
                        graph.add_edge(pnode,endnode,paths=graph[pnode][node][0]['paths'],ofrom=graph[pnode][node][0]['ofrom'],oto=graph[pnode][node][0]['oto'])
                    else:
                        if not graph.has_edge(pnode,endnode):
                            graph.add_edge(pnode,endnode,paths=graph[pnode][node]['paths'],ofrom=graph[pnode][node]['ofrom'],oto=graph[pnode][node]['oto'])
                        else:
                            for p in graph[pnode][node]['paths']:
                                graph[pnode][endnode]['paths'].add(p)
        
        if len(startmerge)>0:
            startnode=uuid.uuid4().hex
            graph.add_node(startnode,offsets={},seq="",endpoint=True) #add dummy node for start of each sequence in the subgraph
            graph.graph['startnodes'].append(startnode)

            for node in startmerge: #copy offset values
                for k in graph.node[node]['offsets']:
                    graph.node[startnode]['offsets'][k]=graph.node[node]['offsets'][k]
                #reconnect
                pred=set()
                predids=set()
                for nnode in graph.successors(node):
                    if type(graph)==nx.MultiDiGraph:
                        graph.add_edge(startnode,nnode,paths=graph[node][nnode][0]['paths'],ofrom=graph[node][nnode][0]['ofrom'],oto=graph[node][nnode][0]['oto'])
                    else:
                        if not graph.has_edge(startnode,nnode):
                            graph.add_edge(startnode,nnode,paths=graph[node][nnode]['paths'],ofrom=graph[node][nnode]['ofrom'],oto=graph[node][nnode]['oto'])
                        else:
                            for p in graph[node][nnode]['paths']:
                                graph[startnode][nnode]['paths'].add(p)

        graph.remove_nodes_from(list(startmerge)+list(endmerge))
        logging.debug("Done.")

    if revcomp:
        genome2length=dict()
        #relabel the offsets, determine the length of all genomes in the graph, then l-pos
        for sample in graph.graph['paths']:
            maxp=0
            for node,data in graph.nodes(data=True):
                if graph.graph['path2id'][sample] in data['offsets']:
                    if data['offsets'][graph.graph['path2id'][sample]]+ (node[1]-node[0]) >maxp:
                        maxp=data['offsets'][graph.graph['path2id'][sample]]+(node[1]-node[0])
            genome2length[sample]=maxp
        
        for sample in graph.graph['paths']:
            for node,data in graph.nodes(data=True):
                if graph.graph['path2id'][sample] in data['offsets']:
                    graph.node[node]['offsets'][graph.graph['path2id'][sample]]=genome2length[sample]-(graph.node[node]['offsets'][graph.graph['path2id'][sample]]+(node[1]-node[0]))
        
        graph.reverse(copy=False)

#simply write sequence without the graph topology
def write_fasta(G,T,outputfile="reference.fasta"):
    f=open(outputfile,'wb')
    for i,node in enumerate(nx.topological_sort(G)):
        if isinstance(node,str):
            continue
        i+=1
        data=G.node[node]
        seq=""
        if len(node)==3:
            nodename=node[2]
        else:
            nodename=str(node)
        if 'seq' in data:
            f.write(">%s\n"%nodename)
            f.write(data['seq'].upper()+"\n")
        else:
            if isinstance(node,Interval):
                f.write(">%s\n"%nodename)
                f.write(T[node.begin:node.end].upper()+"\n")
            else:
                f.write(">%s\n"%nodename)
                logging.warn("No sequence for node: %s"%nodename)
    f.close()

def write_gfa(G,T,outputfile="reference.gfa", paths=True, remap=True, toupper=False):
    
    if not outputfile.endswith(".gfa"):
        outputfile+=".gfa"
    
    f=open(outputfile,'wb')
    sep=';'
    f.write('H\tVN:Z:1.0\tCL:Z:%s\n'%" ".join(sys.argv))
    
    sample2id=dict()
    sample2id=G.graph['path2id']

    mapping={}
    
    if type(G)==nx.DiGraph or type(G)==nx.classes.graphviews.SubDiGraph:
        iterator=nx.topological_sort(G)
        logging.debug("Writing gfa in topological order.")
    elif type(G)==nx.MultiDiGraph or type(G)==nx.classes.graphviews.SubMultiDiGraph:
        iterator=G.nodes()
        logging.debug("Writing gfa in random order.")
    else:
        logging.fatal("Unsupported graph type: %s"%type(G))
        sys.exit(1)

    iterator=[node for node in iterator if type(node)!=str] #exclude start/end node

    if remap:
        for i,node in enumerate(iterator): #iterate once to get a mapping of ids to intervals
            mapping[node]=i+1
    else:
        for node in iterator: #quick and dirty..
            mapping[node]=node

    for i,node in enumerate(iterator):

        i+=1
        data=G.node[node]
        seq=""

        if 'seq' in data:
            if toupper:
                f.write('S\t'+str(mapping[node])+'\t'+data['seq'].upper())
            else:
                f.write('S\t'+str(mapping[node])+'\t'+data['seq'])
            seq=data['seq']
        else:
            if isinstance(node,Interval):
                if toupper:
                    seq=T[node.begin:node.end].upper()
                else:
                    seq=T[node.begin:node.end]
                f.write('S\t'+str(mapping[node])+'\t'+seq)
            elif isinstance(node,tuple):
                if toupper:
                    seq=T[node[0]:node[0]+G.node[node]['l']].upper()
                else:
                    seq=T[node[0]:node[0]+G.node[node]['l']]
                f.write('S\t'+str(mapping[node])+'\t'+seq)
            else:
                logging.error("Node type unknown: %s"%node)
                f.write('S\t'+str(mapping[node])+'\t')
        
        f.write("\n")
        
        for to in G[node]:
            if type(to)==str:
                continue
            
            if type(G)==nx.MultiDiGraph:
                for edgeid in G[node][to]:
                    if 'cigar' in G[node][to][edgeid]:
                        cigar=G[node][to][edgeid]['cigar']
                    f.write("L\t"+str(mapping[node])+"\t"+G[node][to][edgeid]['ofrom']+"\t"+str(mapping[to])+"\t"+G[node][to][edgeid]['oto']+"\t"+(G[node][to][edgeid]['cigar'] if 'cigar' in G[node][to][edgeid] else "0M")+"\n")
            else:
                if 'ofrom' in G[node][to] and 'oto' in G[node][to]:
                    f.write("L\t"+str(mapping[node])+"\t"+G[node][to]['ofrom']+"\t"+str(mapping[to])+"\t"+G[node][to]['oto']+"\t"+(G[node][to]['cigar'] if 'cigar' in G[node][to] else "0M")+"\n")
                else: #if not there, assume same orientation
                    f.write("L\t"+str(mapping[node])+"\t+\t"+str(mapping[to])+"\t+\t"+(G[node][to]['cigar'] if 'cigar' in G[node][to] else "0M")+"\n")

    #write paths
    # for sample in G.graph['paths']:
    for sample in G.graph['path2id']:
        sid=G.graph['path2id'][sample]
        logging.debug("Writing path: %s (sid=%d)"%(sample,sid))
        path=[]
        cigarpath=[]
        logging.debug("Startnodes in graph: %s"%G.graph['startnodes'])
        logging.debug("Endnodes in graph: %s"%G.graph['endnodes'])
        for node in G.graph['startnodes']:
            if node in G: #might be a subgraph of the actual graph
                if sid in G.node[node]['offsets']: #found the start of this path
                    if type(node)!=str: #skip nodes that mark begin/end of an alignment
                        path.append(str(mapping[node])+'+')
                        cigarpath.append("0M")
                    while True:
                        oute=[(u,v,d) for u,v,d in G.out_edges(node,data=True) if sid in d['paths']]
                        # logging.debug("node: %s oute: %s"%(node,str(oute)))
                        if len(oute)==0:
                            logging.warn("Path: \"%s\" (sid=%s) doesnt reach end node, stops at %s!"%(sample,sid,node))
                            break
                        elif len(oute)>1:
                            logging.error("Ambiguity in path for: %s at node: %s"%(sample,node))
                            break
                        else:
                            v=oute[0][1]
                            d=oute[0][2]
                            # if type(v)==str: #end node
                                # break
                            if v in G.graph['endnodes']:
                                break
                            elif type(v)==str: #skip nodes that mark begin/end of an alignment, but dont stop traversing
                                node=v
                                continue
                            else:
                                path.append(str(mapping[v]) + (d['oto'] if 'oto' in d else '+'))
                                if type(node)!=str:
                                    cigarpath.append(d['cigar'] if 'cigar' in d else "0M")
                                node=v
                    break
            else:
                logging.debug("Startnode not in the specified graph!")

        f.write("P\t"+sample+"\t"+",".join(path)+"\t"+",".join(cigarpath)+"\n")
    
    f.close()

def write_gml(G,T,outputfile="reference",partition=False,hwm=4000):
    G=G.copy()
    mapping={}

    if 'paths' in G.graph:
        totn=len(G.graph['paths'])
        logging.debug("Graph contains %d samples"%totn)
    else:
        totn=0
    
    for key in G.graph:
        G.graph[key]=str(G.graph[key])
    
    if type(G)==nx.MultiDiGraph:
        for n1,n2,k,d in G.edges(keys=True,data=True):
            for key in d:
                v=d[key]
                if not isinstance(v,str) and not isinstance(v,int):
                    G[n1][n2][k][key]=str(v)
    else:
        for n1,n2,d in G.edges(data=True):
            for key in d:
                v=d[key]
                if not isinstance(v,str) and not isinstance(v,int):
                    G[n1][n2][key]=str(v)

    for n,d in G.nodes(data=True):
        mapping[n]=str(n)
        d=G.node[n]
        
        if 'offsets' in d:
            G.node[n]['n']=len(d['offsets'])
        
        for key in d:
            v=d[key]

            if type(v)!=str and type(v)!=int:
                G.node[n][key]=str(v)
        
        if 'seq' not in G.node[n]:
            if isinstance(n,Interval):
                G.node[n]['seq']=T[n.begin:n.end].upper()
            else:
                G.node[n]['seq']=""
        G.node[n]['l']=len(G.node[n]['seq'])
        # G.node[n]['seqstart']=G.node[n]['seq'][:20]
        G.node[n]['seqend']=G.node[n]['seq'][-20:]
    
    G=nx.relabel_nodes(G,mapping)

    outputfiles=[]
    
    if partition:
        logging.debug("Trying to partion graph into subgraphs of size %d."%hwm)
        i=0

        # for sgi,subset in enumerate(nx.connected_components(G.to_undirected())):
        for sgi,subset in enumerate(nx.weakly_connected_component_subgraphs(G)):
            logging.debug("Partitioning connected component: %d"%sgi)
            sgn=[]
            g=G.subgraph(subset)
            gn=G.graph['paths']
            for n in nx.topological_sort(g):
                sgn.append(n)
                if G.node[n]['n']==totn: #join/split node
                    logging.debug("Can split graph at node: %s."%n)
                    if len(sgn)>=hwm:
                        logging.debug("Splitting graph at node: %s"%n)
                        sg=G.subgraph(sgn)
                        fn=outputfile+'.'+str(i)+'.gml'
                        nx.write_gml(sg,fn)
                        outputfiles.append(fn)
                        sgn=[n]
                        i+=1
            
            if len(sgn)>1:
                sg=G.subgraph(sgn)
                fn=outputfile+'.'+str(i)+'.gml'
                nx.write_gml(sg,fn)
                i+=1
                outputfiles.append(fn)
    else:
        if not outputfile.endswith(".gml"):
            outputfile=outputfile+'.gml'
        nx.write_gml(G,outputfile)
        outputfiles.append(outputfile)
    
    return outputfiles

def mum_kdtree(mums, depth=0):
    n=len(mums)
    if n==0:
        return None
    if n==1:
        return mums[0]
    k=len(mums[0][1]) #k-tuple with startpositions
    splitdim=depth % k
    smums=sorted(mums,key=lambda m: m[1][splitdim])
    splitvalue=smums[n/2][1][splitdim] #take median for splitting
    if splitvalue==smums[0][1][splitdim]:
        splitvalue+=1
    left=[p for p in smums if p[1][splitdim] < splitvalue]
    right=[p for p in smums if p[1][splitdim] >= splitvalue]
    return { 'left': mum_kdtree(left, depth=depth+1) , 'split' : splitvalue, 'right': mum_kdtree(right, depth=depth+1) }

def mum_range_search(kdtree, qstart, qend):
    k=len(qstart)
    points=[]
    stack=[(kdtree,0)]
    while len(stack)!=0:
        tree,depth=stack.pop()
        splitdim=depth%k

        if tree==None: #something is wrong
            continue

        if isinstance(tree,tuple): #reached leaf, tree==point
            if tree[1][splitdim]>=qstart[splitdim] and tree[1][splitdim]<=qend[splitdim]:
                #check ik point is contained in the range query
                for d in range(k):
                    if tree[1][d]>=qstart[d] and tree[1][d]<=qend[d]:
                        continue
                    else:
                        break
                else:
                    points.append(tree)
            continue

        splitvalue=tree['split']
        
        if splitvalue>=qstart[splitdim] and splitvalue<=qend[splitdim]: #intersect
            if splitvalue != qstart[splitdim]: #equal values go right
                stack.append((tree['left'],depth+1))
            stack.append((tree['right'],depth+1))
        elif splitvalue < qstart[splitdim]:
            stack.append((tree['right'],depth+1))
        else:
            stack.append((tree['left'],depth+1))
    return points



def kdtree(points, k, depth=0):
    n=len(points)
    if n==0:
        return None
    if n==1:
        return points[0]
    splitdim=depth % k
    spoints=sorted(points,key=lambda p: p[splitdim])
    splitvalue=spoints[n/2][splitdim] #take median for splitting
    if splitvalue==spoints[0][splitdim]:
        splitvalue+=1
    left=[p for p in spoints if p[splitdim] < splitvalue]
    right=[p for p in spoints if p[splitdim] >= splitvalue]
    return { 'left': kdtree(left, k, depth=depth+1) , 'split' : splitvalue, 'right': kdtree(right, k, depth=depth+1) }

#return all points within the query range
def range_search(kdtree, qstart, qend):
    k=len(qstart)
    points=[]
    stack=[(kdtree,0)]
    while len(stack)!=0:
        tree,depth=stack.pop()
        splitdim=depth%k

        if tree==None: #something is wrong
            continue

        if isinstance(tree,tuple): #reached leaf, tree==point
            if tree[splitdim]>=qstart[splitdim] and tree[splitdim]<=qend[splitdim]:
                #check ik point is contained in the range query
                for d in range(k):
                    if tree[d]>=qstart[d] and tree[d]<=qend[d]:
                        continue
                    else:
                        break
                else:
                    points.append(tree)
            continue
        
        splitvalue=tree['split']

        if splitvalue>=qstart[splitdim] and splitvalue<=qend[splitdim]: #intersect
            if splitvalue != qstart[splitdim]: #equal values go right
                stack.append((tree['left'],depth+1))
            stack.append((tree['right'],depth+1))
        elif splitvalue < qstart[splitdim]:
            stack.append((tree['right'],depth+1))
        else:
            stack.append((tree['left'],depth+1))

    return points

#copy interval based nodes to sequence attribute
def seq2node(G,T,toupper=True,remap=False):
    i=1
    mapping=dict()
    for node in G:
        if isinstance(node,Interval):
            if toupper and G.node[node]['aligned']>0:
                G.node[node]['seq']=T[node.begin:node.end].upper()
            else:
                G.node[node]['seq']=T[node.begin:node.end]
            if remap:
                mapping[node]=i
                i+=1
    if remap: #get rid of interval objects
        G=nx.relabel_nodes(G,mapping,copy=False)

#converts a list of aligned sequences to a graph
def aln2graph(seqs,names,idoffset=0,confidence=None,minconf=0,path2id=None,offsets=None,addstartstop=True):
    nn=idoffset
    ng=nx.DiGraph()

    if path2id!=None:
        for name in names:
            assert(name in path2id)
        ng.graph['paths']=names
        ng.graph['path2id']=path2id
        ng.graph['id2path']={path2id[p]:p for p in path2id}
    else:
        ng.graph['paths']=[]
        ng.graph['path2id']=dict()
        ng.graph['id2path']=dict()

        for name,seq in zip(names,seqs):
            sid=len(ng.graph['paths'])
            ng.graph['path2id'][name]=sid
            ng.graph['id2path'][sid]=name
            ng.graph['paths'].append(name)

    if confidence==None:
        confidence=[100]*len(seqs[0])

    # offsets={o:-1 for o in range(len(seqs))}

    if offsets==None:
        offsets={ng.graph['path2id'][p]:-1 for p in ng.graph['path2id']}
    else:
        offsets={sid:offsets[sid]-1 for sid in offsets}

    nid=nn

    for i in xrange(len(seqs[0])):
        col={}
        base2node={}
        sid2node={}
        p=confidence[i]

        for j in xrange(len(seqs)):
            pid=ng.graph['path2id'][names[j]]

            if seqs[j][i] in col:
                # col[seqs[j][i]].add(j)
                col[seqs[j][i]].add(pid)
            else:
                # col[seqs[j][i]]=set([j])
                col[seqs[j][i]]=set([pid])

            if seqs[j][i]!='-':
                # offsets[j]+=1
                offsets[pid]+=1

        for base in col:
            if i==0:
                assert(len(col[base])>0)
                # if len(col[base])>0:
                if p>=minconf:
                    ng.add_node(nid,seq=base,offsets={sid:offsets[sid] for sid in offsets if sid in col[base]},p=[p])
                    base2node[base]=nid
                    for sid in col[base]:
                        assert(sid not in sid2node)
                        sid2node[sid]=nid
                    nid+=1
                else: #new node per sequence
                    for sid in col[base]:
                        ng.add_node(nid,seq=base,offsets={sid:offsets[sid]},p=[p])
                        assert(sid not in sid2node)
                        sid2node[sid]=nid
                        if base in base2node:
                            base2node[base].append(nid)
                        else:
                            base2node[base]=[nid]
                        nid+=1
            else:

                if p>=minconf and pp>=minconf:
                    for pbase in pcol:
                        overlap=pcol[pbase].intersection(col[base])
                        if len(overlap)==0:
                            continue
                        elif len(overlap)==len(col[base])==len(pcol[pbase]): #append seq
                            ng.node[pbase2node[pbase]]['seq']+=base
                            ng.node[pbase2node[pbase]]['p']+=[p]
                            
                            base2node[base]=pbase2node[pbase]
                            
                            for sid in overlap:
                                assert(sid not in sid2node)
                                sid2node[sid]=sid2pnode[sid]
                        else:
                            assert(len(overlap)>0)
                            if base not in base2node: #if not already there
                                ng.add_node(nid,seq=base,offsets=dict(),p=[p])
                                base2node[base]=nid
                                for sid in col[base]:
                                    assert(sid not in sid2node)
                                    sid2node[sid]=nid
                                nid+=1
                            for sid in overlap:
                                ng.node[base2node[base]]['offsets'][sid]=offsets[sid]

                            ng.add_edge(pbase2node[pbase],base2node[base],paths=overlap,oto='+',ofrom='+')

                elif p<minconf and pp>=minconf:

                    for sid in col[base]:
                        ng.add_node(nid,seq=base,offsets={sid:offsets[sid]},p=[p])
                        ng.add_edge(sid2pnode[sid],nid,paths={sid},oto='+',ofrom='+')
                        sid2node[sid]=nid
                        if base in base2node:
                            base2node[base].append(nid)
                        else:
                            base2node[base]=[nid]
                        nid+=1

                elif p>=minconf and pp<minconf:
                    
                    ng.add_node(nid,seq=base,offsets=dict(),p=[p])
                    for sid in col[base]:
                        ng.node[nid]['offsets'][sid]=offsets[sid]
                        if not ng.has_edge(sid2pnode[sid],nid):
                            ng.add_edge(sid2pnode[sid],nid,paths={sid},oto='+',ofrom='+')
                        else:
                            ng[sid2pnode[sid]][nid]['paths'].add(sid)
                        sid2node[sid]=nid
                        base2node[base]=nid
                    nid+=1

                elif p<minconf and pp<minconf:
                    for sid in col[base]:
                        ng.node[sid2pnode[sid]]['seq']+=base
                        ng.node[sid2pnode[sid]]['p'].append(p)
                    sid2node=sid2pnode

                else:
                    logging.error("Impossible combination!")
                    sys.exit(1)

        assert(len(sid2node)==len(seqs))
        sid2pnode=sid2node
        pbase2node=base2node
        pcol=col
        pp=p

    #remove gaps from graph
    remove=[]
    for node,data in ng.nodes(data=True):
        incroffset=False
        if data['seq'][0]=='-':
            incroffset=True

        data['seq']=data['seq'].replace("-","")
        if data['seq']=="":
            remove.append(node)
        elif incroffset:
            for sid in data['offsets']:
                data['offsets'][sid]+=1

        if len(data['offsets'])>1:
            data['aligned']=1
        else:
            data['aligned']=0

    for node in remove:
        ine=ng.in_edges(node,data=True)
        oute=ng.out_edges(node,data=True)
        for in1,in2,ind in ine:
            for out1,out2,outd in oute:
                overlap=ind['paths'].intersection(outd['paths'])
                if len(overlap)>=1:
                    if ng.has_edge(in1,out2):
                        ng[in1][out2]['paths']=ng[in1][out2]['paths'] | overlap
                    else:
                        ng.add_edge(in1,out2,paths=overlap,ofrom='+',oto='+')

    ng.remove_nodes_from(remove)

    #TODO: use utils.contract
    #contract edges
    updated=True
    while updated:
        updated=False
        for v,t in ng.edges():
            if ng.out_degree(v)==ng.in_degree(t)==1:
                if ng.node[v]['offsets'].keys()==ng.node[t]['offsets'].keys():
                    ng.node[v]['seq']+=ng.node[t]['seq']
                    for suc in ng.successors(t):
                        ng.add_edge(v,suc,**ng[t][suc])
                    ng.remove_node(t)
                    updated=True
                    break

    path2start=dict()
    path2end=dict()

    #determine start and end for each path
    for node,data in ng.nodes(data=True):
        for sid in data['offsets']:
            if sid not in path2start or data['offsets'][sid]<path2start[sid][1]:
                path2start[sid]=(node,data['offsets'][sid])

        for sid in data['offsets']:
            if sid not in path2end or data['offsets'][sid]>path2end[sid][1]:
                path2end[sid]=(node,data['offsets'][sid])

    if addstartstop:
        start=uuid.uuid4().hex
        ng.add_node(start,offsets=dict(),endpoint=True)
        ng.graph['startnodes']=[start]
        for sid in path2start:
            v=start
            t=path2start[sid][0]
            ng.node[start]['offsets'][sid]=path2start[sid][1]
            if ng.has_edge(v,t):
                ng[v][t]['paths'].add(sid)
            else:
                ng.add_edge(v,t,paths=set([sid]),ofrom="+",oto="+")

        end=uuid.uuid4().hex
        ng.add_node(end,offsets=dict(),endpoint=True)
        ng.graph['endnodes']=[end]
        for sid in path2end:
            v=path2end[sid][0]
            t=end
            ng.node[end]['offsets'][sid]=path2end[sid][1]+len(ng.node[path2end[sid][0]]['seq'])
            if ng.has_edge(v,t):
                ng[v][t]['paths'].add(sid)
            else:
                ng.add_edge(v,t,paths=set([sid]),ofrom="+",oto="+")

    return ng,nid


import utils
import bubbles
import networkx as nx
import os
import logging

def unzip(args):
    if not args.graph[0].endswith(".gfa"):
        logging.fatal("Invalid gfa file.")
        return

    # G=nx.MultiDiGraph()
    G=nx.DiGraph()
    utils.read_gfa(args.graph[0], None, None, G, remap=False)

    if args.source==None and args.sink==None:
        unzip_graph(G,args,minunzip=args.minunzip)
    else:
        b=bubbles.Bubble(G,args.source,args.sink)
        unzip_bubble(G,b,minunzip=args.minunzip,idoffset=max([n for n in G.nodes() if type(n)==int])+1)

    if args.output==None:
        of=os.path.splitext(args.graph[0])[0]+".unzipped.gfa"
    else:
        of=args.output+".gfa"

    utils.write_gfa(G,None,outputfile=of)

#determine uncertainty about bubble positions
def unzip_graph(G,args,minunzip=0):
    nid=max([n for n in G.nodes() if type(n)==int])
    nid+=1

    for b in bubbles.bubbles(G):
        
        if b.maxsize-b.minsize<args.mindiff:
            logging.debug("Skipping bubble %s, diff between smallest and largest allele (%dbp) is smaller than mindiff=%d."%(str(b.nodes),b.maxsize-b.minsize,args.mindiff))
            continue

        if args.maxdiff and b.maxsize-b.minsize>args.maxdiff:
            logging.debug("Skipping bubble %s, diff between smallest and largest allele (%dbp) is larger than maxdiff=%d."%(str(b.nodes),b.maxsize-b.minsize,args.maxdiff))
            continue

        if isinstance(b,bubbles.Bubble):
            nid=unzip_bubble(G,b,minunzip=minunzip,idoffset=nid)

def unzip_bubble(G,b,minunzip=0,idoffset=0):
    
    wiggle=b.getwiggle(minwiggle=minunzip)

    if type(b.sink)==str:
        wiggle=(wiggle[0],0)

    if type(b.source)==str:
        wiggle=(0,wiggle[1])

    if wiggle!=(0,0):
        logging.debug("Unzipping bubble between %s and %s"%(b.source,b.sink))
        srcl=len(G.node[b.source]['seq'])
        snkl=len(G.node[b.sink]['seq'])
        maxlw=int(round((srcl-2 if srcl>2 else 0)/float(2)))
        maxrw=int(round((snkl-2 if snkl>2 else 0)/float(2)))

        if wiggle[0]>maxlw:
            wiggle=(maxlw,wiggle[1])

        if wiggle[1]>maxrw:
            wiggle=(wiggle[0],maxrw)

        if wiggle[0]>0:
            ls=G.node[b.source]['seq'][-wiggle[0]:]
            assert(G.node[b.source]['seq'][:-wiggle[0]]!="")
            G.node[b.source]['seq']=G.node[b.source]['seq'][:-wiggle[0]]
        else:
            ls=""
        
        if wiggle[1]>0:
            rs=G.node[b.sink]['seq'][:wiggle[1]]
            assert(G.node[b.sink]['seq'][wiggle[1]:]!="")
            G.node[b.sink]['seq']=G.node[b.sink]['seq'][wiggle[1]:]
            G.node[b.sink]['offsets']={k:G.node[b.sink]['offsets'][k]+len(rs) for k in G.node[b.sink]['offsets']}
        else:
            rs=""

        successors=list(G.successors(b.source))
        predecessors=list(G.predecessors(b.sink))

        if ls!="":
            for n in successors:
                if len(list(G.predecessors(n)))>1:
                    # G.add_node(idoffset,seq=ls if n!=b.sink else ls+rs,offsets={p:(G.node[b.source]['offsets'][p]+srcl)-len(ls) for p in G[b.source][n].values()[0]['paths']})
                    G.add_node(idoffset,seq=ls if n!=b.sink else ls+rs,offsets={p:(G.node[b.source]['offsets'][p]+srcl)-len(ls) for p in G[b.source][n]['paths']})
                    # props=G[b.source][n].values()[0].copy() #TODO: consider possibilty of structural variant paths!
                    props=G[b.source][n]
                    G.remove_edge(b.source,n)
                    G.add_edge(b.source,idoffset,**props)
                    G.add_edge(idoffset,n,**props)
                    idoffset+=1
                else:
                    G.node[n]['seq']=ls+G.node[n]['seq']
                    G.node[n]['offsets']={k:G.node[n]['offsets'][k]-len(ls) for k in G.node[n]['offsets']}

        if rs!="":
            for n in predecessors:
                if n==b.source and ls!="":
                    continue #was already handled by looping over successors
                if len(list(G.successors(n)))>1:
                    # G.add_node(idoffset,seq=rs if n!=b.source else ls+rs,offsets={p:(G.node[b.sink]['offsets'][p])-len(rs) for p in G[n][b.sink].values()[0]['paths']})
                    G.add_node(idoffset,seq=rs if n!=b.source else ls+rs,offsets={p:(G.node[b.sink]['offsets'][p])-len(rs) for p in G[n][b.sink]['paths']})
                    # props=G[n][b.sink].values()[0].copy() #TODO: consider possibilty of structural variant paths!
                    props=G[n][b.sink]
                    G.remove_edge(n,b.sink)
                    G.add_edge(n,idoffset,**props)
                    G.add_edge(idoffset,b.sink,**props)
                    idoffset+=1
                else:
                    G.node[n]['seq']=G.node[n]['seq']+rs

    return idoffset



from utils import *

def merge_cmd(args):
    if len(args.graphs)<2:
        logging.fatal("Specify multiple gfa files to merge them.")
        return
    
    G=nx.DiGraph()
    for graph in args.graphs:
        logging.info("Adding %s ..." %graph)
        read_gfa(graph,None,"",G,remap=True)
    
    if args.outprefix!=None:
        write_gfa(G,"",outputfile=args.outprefix+".gfa")
    else:
        write_gfa(G,"",outputfile="_".join([os.path.basename(f)[:os.path.basename(f).rfind('.')] for f in args.graphs])+".gfa")



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 13 17:59:26 2015

@author: jasperlinthorst
"""

from intervaltree import IntervalTree, Interval
import networkx as nx
import sys
import math
import logging
import utils
import traceback
from utils import mem2mums
import math

# from matplotlib import pyplot as plt

def chain(mums,left,right,gcmodel="sumofpairs"):
    if len(mums)==0:
        return []

    logging.debug("Number of anchors before chaining: %d",len(mums))
    
    #use one coordinate system for sorting
    ref=mums[0][2].keys()[0]
    # logging.trace("Ref is %s"%ref)
    mums.append(right)
    mums.sort(key=lambda mum: mum[2][ref]) #sort by reference dimension

    sp2mum=dict()
    for mum in mums:
        sp2mum[mum[2][ref]]=mum

    minscore=-1*utils.gapcost([left[2][k] for k in right[2]],[right[2][k] for k in right[2]])
    logging.debug("Initial cost is: %d"%minscore)

    start=left[2][ref]
    end=right[2][ref]

    link=dict()
    score=dict({left[2][ref]:0})
    
    active=[left]
    processed=[]

    for mum in mums:
        trace=False
        #active=[ep2mum[ep] for ep in utils.range_search(mumeptree,(0,0),[sp-1 for sp in mum[2]])].sort(key=lambda x: score[x], reverse=True)
        remove=[]
        for pmum in processed:
            for crd in pmum[2]:
                if pmum[2][crd]+pmum[0]>mum[2][crd]:
                    break
            else:
                active.append(pmum)
                remove.append(pmum)

        for r in remove:
            processed.remove(r)

        active.sort(key=lambda x: score[x[2][ref]], reverse=True) #sort active by score decreasing, kind of priority queue
        
        w=None
        for amum in active:
            for crd in amum[2]:
                if amum[2][crd]+amum[0]>mum[2][crd]:
                    break
            else:
                s=score[amum[2][ref]] + (args.wscore*(mum[0]*((mum[1]*(mum[1]-1))/2)))

                if w!=None:
                    if w > s: #as input is sorted by score
                        break

                penalty=utils.gapcost([amum[2][k]+amum[0] for k in mum[2]],[mum[2][k] for k in mum[2]],model=gcmodel)

                assert(penalty>=0)

                # tmpw=score[amum[2][ref]] + (args.wscore*(mum[0]*((mum[1]*(mum[1]-1))/2))) - (args.wpen*penalty)
                tmpw=s - (args.wpen*penalty)

                if tmpw>w or w==None:
                    logging.trace("mum: %s --> %s = penalty: %d and score at amum: %d, score at mum: %d"%(str(mum),str(amum),penalty,s,tmpw))
                    w=tmpw
                    best=amum

        link[mum[2][ref]]=best[2][ref]

        score[mum[2][ref]]=w

        processed.append(mum)

    logging.debug("Best score is: %d"%score[end])
    logging.trace("Min score is: %d"%minscore)

    #backtrack
    path=[]
    while end!=start:
        path.append((sp2mum[end],score[end]))
        end=link[end]

    return path[1:]

#determine a subset of genomes for which (length * n) is largest
def segment(mums):
    d=dict()
    for mum in mums:
        k=tuple(sorted([gid for gid,sp in mum[2]]))
        if k in d:
            d[k].append(mum)
        else:
            d[k]=[mum]

    best=0
    for part in d:
        z=sum([m[0] for m in d[part]])*len(part)
        if z>best:
            best=z
            partition=part

    logging.debug("Splitting input genomes: %s"%str(partition))
    return d[partition]

def lookup(mum):
    l,mmn,spd=mum
    if isinstance(spd,dict):
        sp=spd.values()
    elif isinstance(spd,tuple):
        sp=[sp for gid,sp in spd]
    else:
        logging.fatal("Unknown format: %s"%str(spd))

    n=0
    qlpoint=dict()
    qrpoint=dict()
    for pos in sp:
        t=ts[pos]
        assert(len(t)==1)
        node=iter(t).next()
        ndata=G.node[node]
        nsamples=set([o for o in ndata['offsets'].keys() if not G.graph['id2path'][o].startswith("*")])
        n+=len(nsamples)
        rel=pos-node[0]
        for k in nsamples:
            v=ndata['offsets'][k]+rel
            qlpoint[k]=v
            qrpoint[k]=v+l
    return (l,n,qlpoint)

def maptooffsets(mums):
    mapping=dict()
    relmums=[]
    for mum in mums:
        relmum=lookup(mum)
        relmums.append(relmum)
        mapping[tuple(relmum[2].values())]=mum
    return relmums,mapping

def trim_overlap(mums):
    coords=mums[0][2]
    for coord in range(len(coords)):
        if len(mums)<=1: #by definition no more overlaps
            break

        mums.sort(key=lambda m: (m[2][coord][1],-m[0])) #sort by start position, then -1*size

        #filter the partial matches that are now contained
        mums=[mum for i,mum in enumerate(mums) if (i==0 and mums[i+1][2][coord][1]+mums[i+1][0] > mum[2][coord][1]+mum[0] ) or mums[i-1][2][coord][1]+mums[i-1][0]<mum[2][coord][1]+mum[0]]

        if len(mums)<=1: #by definition no more overlaps
            break

        trimmed=[mums[0]]
        for mum in mums[1:]:
            pmum=trimmed[-1]
            overlap = (pmum[2][coord][1]+pmum[0]) - mum[2][coord][1]
            if overlap>0:
                if pmum[0]-overlap>0:
                    trimmed[-1] = (pmum[0]-overlap, pmum[1], pmum[2])
                else:
                    del trimmed[-1]
                if mum[0]-overlap>0:
                    trimmed.append( (mum[0]-overlap, mum[1], tuple((k,v+overlap) for k,v in mum[2]) ))
            else:
                trimmed.append(mum)

        mums=trimmed

    return mums

args=None
splitchain="largest"
maxdepth=None #stop recursion when max depth is reached

def graphmumpicker(mums,idx,precomputed=False,minlength=0):
    try:
        if len(mums)==0:
            return ()
        
        if not precomputed:
            if maxdepth!=None:
                if idx.depth>maxdepth:
                    return ()

            if args.maxsize!=None:
                rpaths=[p for p in G.graph['paths'] if not p.startswith('*')]

                if idx.leftnode==None:
                    lo={G.graph['path2id'][p]: 0 for p in rpaths}
                else:
                    lo={k: G.node[idx.leftnode]['offsets'][k]+(idx.leftnode[1]-idx.leftnode[0]) for k in G.node[idx.leftnode]['offsets']}
                
                if idx.rightnode==None:
                    ro={G.graph['path2id'][p]: G.graph['id2end'][G.graph['path2id'][p]] for p in rpaths}
                else:
                    ro=G.node[idx.rightnode]['offsets']

                for k in set(lo.keys()) & set(ro.keys()):
                    if ro[k]-lo[k]>args.maxsize:
                        break
                else:
                    return () #no break, so all fragments in bubbles are smaller than maxsize

            logging.debug("Selecting input multimums (for %d samples) out of: %d mums"%(idx.nsamples, len(mums)))
            mmums=[mum for mum in mums if mum[1]==idx.nsamples] #subset only those mums that apply to all indexed genomes/graphs
            
            if len(mmums)==0 and idx.nsamples>2:
                logging.debug("No MUMS that span all input genomes, segment genomes.")
                mmums=segment(mums)
                logging.debug("Segmented genomes/graphs into %s, now %d MUMS for chaining."%(mmums[0][2],len(mmums)))

            if args.trim:
                logging.debug("Trimming overlap between mums.")
                mmums=trim_overlap(mmums)
                if len(mmums)==0:
                    return ()

            mmums.sort(key=lambda mum: mum[0], reverse=True) #sort by size

            logging.debug("Mapping indexed positions to relative postions within genomes.")

            relmums,mapping=maptooffsets(mmums) #and convert tuple to dict for fast lookup in chaining

            logging.debug("Subset to same group of samples")
            relmums.sort(key=lambda m: (m[1],m[0])) #sort by n, than l
            relmums=[mum for mum in relmums if mum[2].keys()==relmums[-1][2].keys()] #subset to only those mums that apply to the same set
            
            logging.debug("Left with %d mums"%len(relmums))

            if idx.leftnode!=None:
                spd=dict()
                for k in relmums[-1][2].keys():
                    spd[k]=G.node[idx.leftnode]['offsets'][k]+(idx.leftnode[1]-idx.leftnode[0])-1
                left=(0,0,spd)
            else:
                spd=dict()
                for sid in relmums[-1][2].keys():
                    spd[sid]=-1
                left=(0,0,spd)

            if idx.rightnode!=None:
                spd=dict()
                for k in relmums[-1][2].keys():
                    spd[k]=G.node[idx.rightnode]['offsets'][k]
                right=(0,0,spd)
            else:
                spd=dict()
                for sid in relmums[-1][2].keys():
                    spd[sid]=G.graph['id2end'][sid]
                right=(0,0,spd)

            # if minlength==0: #autodetermine significant subset
                # relmums=[mum for mum in relmums if 1-((1-((.25**(mum[1]-1))**mum[0]))**o)<pcutoff] #subset to only significant mums
            
            if len(relmums)==0:
                logging.debug("No more significant MUMs.")
                return ()

            skipleft=[]
            skipright=[]

            if len(relmums)==1:
                splitmum=relmums[0]
            else:
                if len(relmums)>args.maxmums:
                    logging.debug("Number of MUMs exceeds cap (%d), taking largest %d"%(len(mmums),args.maxmums))
                    relmums=relmums[-args.maxmums:]

                logging.debug("Chaining %d mums"%len(relmums))
                chainedmums=chain(relmums,left,right,gcmodel=args.gcmodel)[::-1]

                logging.debug("Selected chain of %d mums"%len(chainedmums))
                if len(chainedmums)==0:
                    return ()

                if splitchain=="balanced":
                    logging.debug("Selecting MUM from chain on position within chain.")
                    optsplit=None
                    for mum,score in chainedmums: #determine optimal split in chain
                        lseq=0
                        rseq=0
                        for crd in mum[2]:
                            lseq=mum[2][crd]
                            assert(lseq>=0)
                            rseq=right[2][crd]-mum[2][crd]+mum[0]
                            assert(rseq>=0)
                        if optsplit==None or abs(lseq-rseq)<optsplit:
                            optsplit=abs(lseq-rseq)
                            splitmum=mum
                elif splitchain=="largest":
                    logging.debug("Selecting MUM from chain based on size.")
                    splitmum=sorted(chainedmums,key=lambda m:m[0][0])[-1][0]
                else: #select at random
                    logging.debug("Selecting MUM from chain at random.")
                    splitmum=chainedmums[random.randint(0,len(chainedmums)-1)][0]

                if args.seedsize>0:
                    t=skipleft
                    scoreatsplit=0
                    for mum,score in chainedmums:
                        if mum==splitmum:
                            scoreatsplit=score
                            t=skipright
                            continue
                        t.append( (mapping[tuple(mum[2].values())], score-scoreatsplit) )
                        # t.append( mapping[tuple(mum[2].values())] )
                    skipleft=[(mum,score) for mum,score in skipleft if mum[0]>=args.seedsize]
                    skipright=[(mum,score) for mum,score in skipright if mum[0]>=args.seedsize]

            splitmum=mapping[tuple(splitmum[2].values())]

            if minlength==0: #experimental, use significance to determine valid anchor length when minlength is set to 0
                o=1
                for p in left[2]:
                    o=o*(right[2][p]-left[2][p])
                l=splitmum[0]
                n=splitmum[1]
                p=((.25**(n-1)))**l #probability of observing this match by random chance
                if p>0:
                    p=1-math.exp(math.log(1-p) * o) #correct for the number of tests we actually did
                if p>args.pcutoff:
                    logging.info("P-value for: %s (n=%d l=%d o=%d) is %.4g"%(str(splitmum),n,l,o,p))
                    return ()
        else:
            logging.debug("Selecting MUM from precomputed chain")
            chainedmums=mums
            splitmum=chainedmums[len(chainedmums)/2][0]
            skipleft=chainedmums[:len(chainedmums)/2]
            skipright=chainedmums[(len(chainedmums)/2)+1:]
        
        logging.debug("Best MUM has length: %d"%splitmum[0])
        
        logging.debug("Skipleft: %d"%len(skipleft))
        logging.debug("Skipright: %d"%len(skipright))

        return splitmum,skipleft,skipright

    except Exception:
        logging.fatal(traceback.format_exc())

def printSA(index,maxline=100,start=0,end=None,fn="sa.txt"):
    sa=index.SA
    lcp=index.LCP
    t=index.T
    #so=index.SO
    if end==None:
        end=len(sa)
    with open(fn,'w') as f:
        f.write("%d\t%d\n"%(len(sa), len(lcp)))
        assert(len(sa)==len(lcp))
        for i in xrange(len(sa)):
            s=sa[i]
            lcpi=lcp[i]

            if i>0 and i<len(sa)-1:
                l1=lcp[i]
                l2=lcp[i+1]
            elif i==len(sa)-1:
                l1=max([lcp[i-1],lcp[i]])
                l2=0
            else:
                l1=0
                l2=lcp[i+1]

            if i>=start and i<=end:
                #f.write("%s\t%s\t%s\n"%(str(s).zfill(8), str(lcpi).zfill(6), t[s:s+maxline].ljust(maxline) if l1<=maxline else t[s:s+maxline]+"..."+t[s+l1-40:s+l1].ljust(maxline) ) )
                f.write("%s\t%s\t%s\t%s\t%s\n"%(str(s).zfill(8), str(lcpi).zfill(6), t[s:s+maxline] ,t[s+l1-maxline:s+l1], t[s+l2-maxline:s+l2] ) )







import sys

if not sys.version_info[0] == 2:
    print("Invalid version of python, use python 2.")
    sys.exit(1)

import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, Extension

setup(name="reveal", author="Jasper Linthorst", author_email="jasper.linthorst@gmail.com", version="0.2.2",
        url="https://github.com/jasperlinthorst/reveal", description="Graph based multi genome aligner",
        install_requires=['intervaltree','networkx==2','pysam','matplotlib'],
        packages = ['reveal'],
        scripts = ['ez_setup.py'],
        test_suite = 'nose.collector',
        tests_require = ['nose'],
        ext_modules=[ \
                
                Extension("reveallib", ["reveallib/reveal.c","reveallib/interface.c","divsufsort/divsufsort.c","divsufsort/utils.c","divsufsort/sssort.c","divsufsort/trsort.c"], \
                                       include_dirs=['reveallib','divsufsort'], \
                                       libraries=['pthread'], \
                                       define_macros=[('HAVE_CONFIG_H',1),('__STDC_CONSTANT_MACROS',1),('__STDC_FORMAT_MACROS',1),('__STDC_LIMIT_MACROS',1)], \
                                       undef_macros=['NDEBUG'] ), \
                
                Extension("reveallib64", ["reveallib/reveal.c","reveallib/interface.c","divsufsort/divsufsort.c","divsufsort/utils.c","divsufsort/sssort.c","divsufsort/trsort.c"], \
                                       include_dirs=['reveallib','divsufsort'], \
                                       libraries=['pthread'], \
                                       define_macros = [('SA64',1),('BUILD_DIVSUFSORT64',1),('HAVE_CONFIG_H',1),('__STDC_CONSTANT_MACROS',1),('__STDC_FORMAT_MACROS',1), ('__STDC_LIMIT_MACROS',1)], \
                                       undef_macros=['NDEBUG'] ), \
                
                Extension("probconslib", ["probcons/Probcons.cc"], \
                                       include_dirs=['probcons'], \
                                       define_macros=[('NumInsertStates',2),('VERSION',1.12)], \
                                       undef_macros=['NDEBUG'] ), \
                ],
        entry_points = {
        'console_scripts': [
            'reveal = reveal.reveal:main'
            ]
         }
        )




import networkx as nx
import utils

def comp(G):
    for node in G.node:
        G.node[node]['seq']=utils.rc(G.node[node]['seq'])
    
    genome2length=dict()
    #relabel the offsets, determine the length of all genomes in the graph, then l-pos
    for sample in G.graph['paths']:
        maxp=0
        for node,data in G.nodes(data=True):
            if sample in data['offsets']:
                if data['offsets'][sample]+len(data['seq'])>maxp:
                    maxp=data['offsets'][sample]+len(data['seq'])
        genome2length[sample]=maxp
    
    for sample in G.graph['paths']:
        for node,data in G.nodes(data=True):
            if sample in data['offsets']:
                G.node[node]['offsets'][sample]=genome2length[sample]-(G.node[node]['offsets'][sample]+len(data['seq']))
    
    G.reverse(copy=False)
    return G

def comp_cmd(args):
    g=nx.DiGraph()
    g.graph['paths']=[]
    utils.read_gfa(args.graph[0],None,None,g,targetsample=None)
    g=comp(g)
    utils.write_gfa(g,"",outputfile=args.graph[0].replace('.gfa','.rc.gfa'), nometa=False)



import networkx as nx
import utils
import sys
import logging
import os
import uuid

def convert(args):
    for graph in args.graphs:
        
        if args.nocycles:
            g=nx.DiGraph()
        else:
            g=nx.MultiDiGraph()

        g.graph['paths']=[]
        g.graph['path2id']=dict()
        g.graph['id2path']=dict()

        if graph.endswith(".gfa"): #gfa to gml/gfa
            utils.read_gfa(graph,None,None,g,minsamples=args.minsamples,
                                 maxsamples=args.maxsamples,
                                 targetsample=args.targetsample,
                                 remap=False)
            if args.type=="gfa":
                fn=graph.replace(".gfa",".rewrite.gfa")
                graph=utils.write_gfa(g,"", outputfile=fn)
                logging.info("gfa graph written to: %s"%fn)
            elif args.type=="gml":
                fn=utils.write_gml(g,"", hwm=args.hwm, outputfile=graph.replace(".gfa",""), partition=args.partition)
                logging.info("gml graph written to: %s"%fn)
            elif args.type=="maf":
                logging.info("Converting graph to maf..")
                graph2maf(g,graph.replace(".gfa",".maf"))
        
        elif graph.endswith(".maf"): #multiple alignment format, convert to graph
            g=maf2graph(graph)
            filename=graph[:graph.rfind(".")]+".gml"
            utils.write_gml(g,"", outputfile=filename)

            filename=graph[:graph.rfind(".")]+".gfa"
            utils.write_gfa(g,"", outputfile=filename)
            logging.debug("gfa graph written to: %s"%filename)

        elif graph.endswith(".fa") or graph.endswith(".fasta") or graph.endswith(".fna"): #assume fasta to gfa
            if args.aligned:
                seqs=[]
                names=[]
                for name,seq in utils.fasta_reader(graph,keepdash=True):
                    seqs.append(seq)
                    names.append(name)
                g,nid=utils.aln2graph(seqs,names)
            else:
                i=0
                start=uuid.uuid4().hex
                end=uuid.uuid4().hex
                g.graph['startnodes']=[start]
                g.graph['endnodes']=[end]
                g.add_node(start,offsets=dict())
                g.add_node(end,offsets=dict())
                for i,v in enumerate(utils.fasta_reader(graph)):
                    name,seq=v
                    g.graph['paths'].append(name)
                    g.graph['path2id'][name]=i
                    g.graph['id2path'][i]=name
                    g.node[start]['offsets'][i]=0
                    g.node[end]['offsets'][i]=len(seq)
                    g.add_node(i,offsets={i:0},seq=seq)
                    g.add_edge(start,i,paths=set([i]))
                    g.add_edge(i,end,paths=set([i]))

            filename=graph[:graph.rfind(".")]+".gfa"
            utils.write_gfa(g,"", outputfile=filename)
            logging.debug("gfa graph written to: %s"%filename)
        else:
            logging.fatal("Unknown filetype, need gfa or fasta extension.")
            return

#converts a multiple alignment format file to a graph
def maf2graph(maffile):
    files=set()
    G=nx.MultiDiGraph()

    startnode=uuid.uuid4().hex
    endnode=uuid.uuid4().hex

    G.graph['startnodes']=set([startnode])
    G.graph['endnodes']=set([endnode])
    G.graph['path2id']=dict()

    G.add_node(startnode,offsets=dict())
    G.add_node(endnode,offsets=dict())

    nid=0
    with open(maffile,"r") as maf:
        for line in maf:
            if line.startswith("#"):
                continue
            elif line.startswith("a"): #start of an aligned segment
                nid+=1
                G.add_node(nid,data=dict())
            elif line.startswith("s"):
                cols=line.rstrip().split()
                if '.' in cols[1]: #TODO: use db parameter to specificy a single mfa file with all sequence
                    file,name=cols[1][:cols[1].find('.')],cols[1][cols[1].find('.')+1:]
                    files.add(file)
                else:
                    file=None #args.db?
                    name=cols[1]

                if name not in G.graph['path2id']:
                    G.graph['path2id'][name]=len(G.graph['path2id'])
                    G.node[startnode]['offsets'][G.graph['path2id'][name]]=0

                G.node[nid]['data'][(file,name)]={'start':int(cols[2]),
                                                  'end':int(cols[2])+int(cols[3]),
                                                  'orientation':cols[4],
                                                  'aln':cols[6]
                                                  }
    nid+=1

    remove=[]
    for node,d in G.nodes(data=True):
        if 'data' in d and len(d['data'])==1: #multiplicity of 1, strictly not an alignment
            remove.append(node)

    G.remove_nodes_from(remove)

    db=dict() #map name to sequence
    for file in files:
        for name,seq in utils.fasta_reader(file+".fasta"): #guess that the original file has a ".fasta" extension
            name=name.split()[0]
            key=(file,name)
            if key in db:
                logging.fatal("Non unique contig-name: %s. quit."%name)
                sys.exit(1)
            else:
                db[key]=seq

    remove=[]

    #for every sequence, check that none of the alignments overlap, otherwise assignment is not 1-1
    for file,name in db:
        seq=db[(file,name)]

        intvs=[]
        for node in G:
            if 'data' in G.node[node]: #does the node represent an aligned segment?
                if (file,name) in G.node[node]['data']:
                    intvs.append((G.node[node]['data'][(file,name)]['start'] , G.node[node]['data'][(file,name)]['end'], node))
        
        intvs.sort() #sort by start position
        pstart=0
        pend=0
        pnode=startnode
        unaligned=[]

        for start,end,node in intvs:
            if start>pend:
                unaligned.append((pend,start))
                G.add_node(nid,intv=(pend,start),seq=seq[pend:start])
                G.add_edge(pnode,nid,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")
                G.add_edge(nid,node,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")
                nid+=1
            elif start<pend:
                logging.fatal("Overlapping alignments for sequence: %s.%s --> (%d,%d) and (%d,%d)."%(file,name,pstart,pend,start,end))
                remove.append(node)
                # sys.exit(1)
            else: #no gap, just connect subsequent intervals
                G.add_edge(pnode,node,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")

            pstart,pend,pnode=start,end,node
        
        if len(seq)!=pend:
            unaligned.append((pend,len(seq)))
            G.add_node(nid,intv=((pend,len(seq))),seq=seq[pend:len(seq)])
            G.add_edge(pnode,nid,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")
            G.add_edge(nid,endnode,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")
            nid+=1
        else:
            G.add_edge(pnode,endnode,paths=set([G.graph['path2id'][name]]),ofrom="+",oto="+")

    G.remove_nodes_from(remove)

    # print "Unaligned segments",unaligned

    alignments=[node for node in G if 'data' in G.node[node]]
    
    for node in alignments: #expand all alignments in the graph

        if 'data' in G.node[node]:
            seqs=[]
            names=[]
            offsets={}

            for file,name in G.node[node]['data']:
                seqs.append(G.node[node]['data'][(file,name)]['aln'])
                offsets[G.graph['path2id'][name]]=G.node[node]['data'][(file,name)]['start']
                names.append(name)

            sg,nid=utils.aln2graph(seqs,names,idoffset=nid,path2id=G.graph['path2id'],offsets=offsets)

            nid+=1

            G.add_nodes_from(sg.nodes(data=True))
            G.add_edges_from(sg.edges(data=True))

            assert(len(sg.graph['startnodes'])==1)
            assert(len(sg.graph['endnodes'])==1)

            sgstart=sg.graph['startnodes'][0]
            sgend=sg.graph['endnodes'][0]

            for v,t,d in G.in_edges(node,data=True):
                G.add_edge(v,sgstart,paths=d['paths'],ofrom="+",oto="+")

            for v,t,d in G.out_edges(node,data=True):
                G.add_edge(sgend,t,paths=d['paths'],ofrom="+",oto="+")

            #hack this in here so we can continue
            G.node[sgstart]['seq']=""
            G.node[sgend]['seq']=""
            nx.relabel_nodes(G,{sgstart: nid, sgend: nid+1},copy=False)

            nid+=2

            G.remove_node(node)

    return G

def graph2maf(G,filename):

    if isinstance(G,nx.MultiDiGraph):
        #TODO: decompose global alignment into local alignments by deconnecting structure edges
        #determine set of structure edges
        orgpaths=set([G.graph['path2id'][p] for p in G.graph['paths'] if p.startswith('*')])
        refpaths=set([G.graph['path2id'][p] for p in G.graph['paths'] if not p.startswith('*')])
        
        es=[]
        for e0,e1,d in G.edges(data=True):
            if len(d['paths'] & refpaths)==0: #edge that exclusively represents structural event
                es.append((e0,e1))

        toremove=es
        G.remove_edges_from(toremove)

    sizes={sid:0 for sid in G.graph['id2path']}
    
    with open(filename,'w') as maf:

        for g in nx.weakly_connected_component_subgraphs(G):
            
            longest=0
            sids=set()
            for node in nx.topological_sort(g):
                if type(node)!=str:
                    go=max([0]+[G.node[pred]['graphoffset']+len(G.node[pred]['seq']) for pred in G.predecessors(node) if type(pred)!=str])
                    G.node[node]['graphoffset']=go
                    
                    if go+len(G.node[node]['seq'])>longest:
                        longest=go+len(G.node[node]['seq'])

                    for k in G.node[node]['offsets']:
                        sids.add(k)
                        if G.node[node]['offsets'][k]+len(G.node[node]['seq'])>sizes[k]:
                            sizes[k]=G.node[node]['offsets'][k]+len(G.node[node]['seq'])
            
            ml=max([len(p) for p in G.graph['paths']])

            maf.write("##maf version=1\n")
            maf.write("a\n")
            for sid in sids:
                path=G.graph['id2path'][sid]
                o=0
                sl=0
                maf.write("s %s %d %d + %-10d "%(path.ljust(ml), 0, sizes[G.graph['path2id'][path]], sizes[G.graph['path2id'][path]]) )
                for node in nx.topological_sort(g):
                    if type(node)!=str and sid in G.node[node]['offsets']:
                        while o<G.node[node]['graphoffset']:
                            maf.write("-")
                            o+=1
                        sl+=len(G.node[node]['seq'].replace("-",""))
                        maf.write("%s"%G.node[node]['seq'])
                        o+=len(G.node[node]['seq'])
                
                maf.write("-"*(longest-o)) #pad with dash so all lines are equally long
                maf.write("\n")
            maf.write("\n")



